from __future__ import annotations

import inspect
import sys
from abc import ABC, abstractmethod
from typing import Any, Literal

from pydantic import ConfigDict, Field, RootModel

from hugr import tys
from hugr.hugr.node_port import (
    NodeIdx,  # noqa: TCH001 # pydantic needs this alias in scope
)
from hugr.utils import deser_it

from . import tys as stys
from .tys import (
    ConfiguredBaseModel,
    ExtensionId,
    FunctionType,
    PolyFuncType,
    SumType,
    Type,
    TypeBound,
    TypeRow,
)
from .tys import (
    classes as tys_classes,
)
from .tys import (
    model_rebuild as tys_model_rebuild,
)


class BaseOp(ABC, ConfiguredBaseModel):
    """Base class for ops that store their node's input/output types."""

    # Parent node index of node the op belongs to, used only at serialization time
    parent: NodeIdx

    def insert_port_types(self, in_types: TypeRow, out_types: TypeRow) -> None:
        """Hook to insert type information from the input and output ports into the
        op.
        """

    def insert_child_dfg_signature(self, inputs: TypeRow, outputs: TypeRow) -> None:
        """Hook to insert type information from a child dataflow graph."""

    def display_name(self) -> str:
        """Name of the op for visualisation."""
        return self.__class__.__name__

    @abstractmethod
    def deserialize(self) -> ops.Op:
        """Deserializes the model into the corresponding Op."""


# ----------------------------------------------------------
# --------------- Module level operations ------------------
# ----------------------------------------------------------


class Module(BaseOp):
    """The root of a module, parent of all other `ModuleOp`s."""

    op: Literal["Module"] = "Module"

    def deserialize(self) -> ops.Module:
        return ops.Module()


class FuncDefn(BaseOp):
    """A function definition. Children nodes are the body of the definition."""

    op: Literal["FuncDefn"] = "FuncDefn"

    name: str
    signature: PolyFuncType
    visibility: tys.Visibility = Field(default="Private")

    def deserialize(self) -> ops.FuncDefn:
        poly_func = self.signature.deserialize()
        return ops.FuncDefn(
            self.name,
            params=poly_func.params,
            inputs=poly_func.body.input,
            _outputs=poly_func.body.output,
            visibility=self.visibility,
        )


class FuncDecl(BaseOp):
    """External function declaration, linked at runtime."""

    op: Literal["FuncDecl"] = "FuncDecl"
    name: str
    signature: PolyFuncType
    visibility: tys.Visibility = Field(default="Public")

    def deserialize(self) -> ops.FuncDecl:
        return ops.FuncDecl(
            self.name, self.signature.deserialize(), visibility=self.visibility
        )


class CustomConst(ConfiguredBaseModel):
    c: str
    v: Any


class BaseValue(ABC, ConfiguredBaseModel):
    @abstractmethod
    def deserialize(self) -> val.Value: ...


class CustomValue(BaseValue):
    """An extension constant value, that can check it is of a given [CustomType]."""

    v: Literal["Extension"] = Field(default="Extension", title="ValueTag")
    typ: Type
    value: CustomConst

    def deserialize(self) -> val.Value:
        return val.Extension(
            name=self.value.c,
            typ=self.typ.deserialize(),
            val=self.value.v,
        )


class FunctionValue(BaseValue):
    """A higher-order function value."""

    v: Literal["Function"] = Field(default="Function", title="ValueTag")
    hugr: str

    def deserialize(self) -> val.Value:
        from hugr.hugr import Hugr

        # pydantic stores the serialized dictionary because of the "Any" annotation
        return val.Function(Hugr.from_str(self.hugr))


class SumValue(BaseValue):
    """A Sum variant.

    For any Sum type where this value meets the type of the variant indicated by the tag
    """

    v: Literal["Sum", "Tuple"] = Field(default="Sum", title="ValueTag")
    tag: int = Field(default=0, title="VariantTag")
    typ: SumType | None = Field(default=None, title="SumType")
    vs: list[Value]
    model_config = ConfigDict(
        json_schema_extra={
            "description": (
                "A Sum variant For any Sum type where this value meets the type "
                "of the variant indicated by the tag."
            ),
        }
    )

    def deserialize(self) -> val.Value:
        if self.typ is None:
            # Backwards compatibility of "Tuple" values
            assert self.tag == 0, "Sum type must be provided if tag is not 0"
            vs = deser_it(v.root for v in self.vs)
            typ = tys.Sum(variant_rows=[[v.type_() for v in vs]])
            return val.Sum(0, typ, vs)
        else:
            return val.Sum(
                self.tag, self.typ.deserialize(), deser_it(v.root for v in self.vs)
            )


class Value(RootModel):
    """A constant Value."""

    root: CustomValue | FunctionValue | SumValue = Field(discriminator="v")

    model_config = ConfigDict(json_schema_extra={"required": ["v"]})

    def deserialize(self) -> val.Value:
        return self.root.deserialize()


class Const(BaseOp):
    """A Const operation definition."""

    op: Literal["Const"] = "Const"
    v: Value = Field()

    def deserialize(self) -> ops.Const:
        return ops.Const(self.v.deserialize())


# -----------------------------------------------
# --------------- BasicBlock types ------------------
# -----------------------------------------------


class DataflowBlock(BaseOp):
    """A CFG basic block node. The signature is that of the internal Dataflow
    graph.
    """

    op: Literal["DataflowBlock"] = "DataflowBlock"
    inputs: TypeRow = Field(default_factory=list)
    other_outputs: TypeRow = Field(default_factory=list)
    sum_rows: list[TypeRow]

    def insert_port_types(self, in_types: TypeRow, out_types: TypeRow) -> None:
        num_cases = len(out_types)
        self.sum_rows = [[] for _ in range(num_cases)]

    def insert_child_dfg_signature(self, inputs: TypeRow, outputs: TypeRow) -> None:
        self.inputs = inputs
        pred = outputs[0].root
        assert isinstance(pred, stys.SumType)
        if isinstance(pred.root, stys.UnitSum):
            self.sum_rows = [[] for _ in range(pred.root.size)]
        else:
            self.sum_rows = []
            for variant in pred.root.rows:
                self.sum_rows.append(variant)
        self.other_outputs = outputs[1:]

        # Needed to avoid random '\n's in the pydantic description

    def deserialize(self) -> ops.DataflowBlock:
        return ops.DataflowBlock(
            inputs=deser_it(self.inputs),
            _sum=tys.Sum([deser_it(r) for r in self.sum_rows]),
            _other_outputs=deser_it(self.other_outputs),
        )

    model_config = ConfigDict(
        json_schema_extra={
            "description": "A CFG basic block node."
            " The signature is that of the internal Dataflow graph.",
        }
    )


class ExitBlock(BaseOp):
    """The single exit node of the CFG, has no children, stores the types of
    the CFG node output.
    """

    op: Literal["ExitBlock"] = "ExitBlock"
    cfg_outputs: TypeRow

    model_config = ConfigDict(
        json_schema_extra={
            # Needed to avoid random '\n's in the pydantic description
            "description": "The single exit node of the CFG, has no children,"
            " stores the types of the CFG node output.",
        }
    )

    def deserialize(self) -> ops.ExitBlock:
        return ops.ExitBlock(deser_it(self.cfg_outputs))


# ---------------------------------------------
# --------------- DataflowOp ------------------
# ---------------------------------------------


class DataflowOp(BaseOp):
    pass


class Input(DataflowOp):
    """An input node. The outputs of this node are the inputs to the parent node."""

    op: Literal["Input"] = "Input"
    types: TypeRow = Field(default_factory=list)

    def insert_port_types(self, in_types: TypeRow, out_types: TypeRow) -> None:
        assert len(in_types) == 0
        self.types = list(out_types)

    def deserialize(self) -> ops.Input:
        return ops.Input(types=[t.deserialize() for t in self.types])


class Output(DataflowOp):
    """An output node. The inputs are the outputs of the function."""

    op: Literal["Output"] = "Output"
    types: TypeRow = Field(default_factory=list)

    def insert_port_types(self, in_types: TypeRow, out_types: TypeRow) -> None:
        assert len(out_types) == 0
        self.types = list(in_types)

    def deserialize(self) -> ops.Output:
        return ops.Output(deser_it(self.types))


class Call(DataflowOp):
    """Call a function directly.

    The first port is connected to the def/declare of the function being called
    directly, with a `ConstE<Graph>` edge. The signature of the remaining ports matches
    the function being called.
    """

    op: Literal["Call"] = "Call"
    func_sig: PolyFuncType
    type_args: list[stys.TypeArg]
    instantiation: FunctionType

    model_config = ConfigDict(
        # Needed to avoid random '\n's in the pydantic description
        json_schema_extra={
            "description": (
                "Operation to call a function directly. The first port is "
                "connected to the def/declare of the function being called directly, "
                "with a `Static<FunctionType>` edge. The signature of the remaining "
                "ports matches the function being called."
            )
        }
    )

    def deserialize(self) -> ops.Call:
        return ops.Call(
            self.func_sig.deserialize(),
            self.instantiation.deserialize(),
            deser_it(self.type_args),
        )


class CallIndirect(DataflowOp):
    """Call a function indirectly.

    Like call, but the first input is a standard dataflow graph type.
    """

    op: Literal["CallIndirect"] = "CallIndirect"
    signature: FunctionType = Field(default_factory=FunctionType.empty)

    def insert_port_types(self, in_types: TypeRow, out_types: TypeRow) -> None:
        fun_ty = in_types[0].root
        assert isinstance(fun_ty, FunctionType)
        assert len(fun_ty.input) == len(in_types) - 1
        assert len(fun_ty.output) == len(out_types)
        self.signature = fun_ty

    def deserialize(self) -> ops.CallIndirect:
        return ops.CallIndirect(self.signature.deserialize())


class LoadConstant(DataflowOp):
    """An operation that loads a static constant in to the local dataflow graph."""

    op: Literal["LoadConstant"] = "LoadConstant"
    datatype: Type

    def deserialize(self) -> ops.LoadConst:
        return ops.LoadConst(self.datatype.deserialize())


class LoadFunction(DataflowOp):
    """Load a static function in to the local dataflow graph."""

    op: Literal["LoadFunction"] = "LoadFunction"
    func_sig: PolyFuncType
    type_args: list[stys.TypeArg]
    instantiation: FunctionType

    def deserialize(self) -> ops.LoadFunc:
        return ops.LoadFunc(
            self.func_sig.deserialize(),
            self.instantiation.deserialize(),
            deser_it(self.type_args),
        )


class DFG(DataflowOp):
    """A simply nested dataflow graph."""

    op: Literal["DFG"] = "DFG"
    signature: FunctionType = Field(default_factory=FunctionType.empty)

    def insert_child_dfg_signature(self, inputs: TypeRow, outputs: TypeRow) -> None:
        self.signature = FunctionType(input=list(inputs), output=list(outputs))

    def deserialize(self) -> ops.DFG:
        sig = self.signature.deserialize()
        return ops.DFG(sig.input, sig.output)


# ------------------------------------------------
# --------------- ControlFlowOp ------------------
# ------------------------------------------------


class Conditional(DataflowOp):
    """Conditional operation, defined by child `Case` nodes for each branch."""

    op: Literal["Conditional"] = "Conditional"
    other_inputs: TypeRow = Field(default_factory=list)  # Remaining input types
    outputs: TypeRow = Field(default_factory=list)  # Output types
    sum_rows: list[TypeRow] = Field(
        description="The possible rows of the Sum input", default_factory=list
    )

    def insert_port_types(self, in_types: TypeRow, out_types: TypeRow) -> None:
        # First port is a predicate, i.e. a sum of tuple types. We need to unpack
        # those into a list of type rows
        pred = in_types[0]
        assert isinstance(pred.root, stys.SumType)
        sum = pred.root.root
        if isinstance(sum, stys.UnitSum):
            self.sum_rows = [[] for _ in range(sum.size)]
        else:
            assert isinstance(sum, stys.GeneralSum)
            self.sum_rows = []
            for ty in sum.rows:
                self.sum_rows.append(ty)
        self.other_inputs = list(in_types[1:])
        self.outputs = list(out_types)

    def deserialize(self) -> ops.Conditional:
        return ops.Conditional(
            tys.Sum([deser_it(r) for r in self.sum_rows]),
            deser_it(self.other_inputs),
            deser_it(self.outputs),
        )


class Case(BaseOp):
    """Case ops - nodes valid inside Conditional nodes."""

    op: Literal["Case"] = "Case"
    # The signature of the contained dataflow graph.
    signature: FunctionType = Field(default_factory=FunctionType.empty)

    def insert_child_dfg_signature(self, inputs: TypeRow, outputs: TypeRow) -> None:
        self.signature = stys.FunctionType(input=list(inputs), output=list(outputs))

    def deserialize(self) -> ops.Case:
        sig = self.signature.deserialize()
        return ops.Case(inputs=sig.input, _outputs=sig.output)


class TailLoop(DataflowOp):
    """Tail-controlled loop."""

    op: Literal["TailLoop"] = "TailLoop"
    # Types that are only input
    just_inputs: TypeRow = Field(default_factory=list)
    # Types that are only output
    just_outputs: TypeRow = Field(default_factory=list)
    # Types that are appended to both input and output:
    rest: TypeRow = Field(default_factory=list)

    def insert_port_types(self, in_types: TypeRow, out_types: TypeRow) -> None:
        assert in_types == out_types
        # self.just_inputs = list(in_types)
        # self.just_outputs = list(out_types)
        self.rest = list(in_types)

    def deserialize(self) -> ops.TailLoop:
        return ops.TailLoop(
            just_inputs=deser_it(self.just_inputs),
            _just_outputs=deser_it(self.just_outputs),
            rest=deser_it(self.rest),
        )


class CFG(DataflowOp):
    """A dataflow node which is defined by a child CFG."""

    op: Literal["CFG"] = "CFG"
    signature: FunctionType = Field(default_factory=FunctionType.empty)

    def insert_port_types(self, inputs: TypeRow, outputs: TypeRow) -> None:
        self.signature = FunctionType(
            input=list(inputs),
            output=list(outputs),
        )

    def deserialize(self) -> ops.CFG:
        sig = self.signature.deserialize()
        return ops.CFG(inputs=sig.input, _outputs=sig.output)


ControlFlowOp = Conditional | TailLoop | CFG


class ExtensionOp(DataflowOp):
    """A user-defined operation that can be downcasted by the extensions that define
    it.
    """

    op: Literal["Extension"] = "Extension"
    extension: ExtensionId
    name: str
    signature: stys.FunctionType = Field(default_factory=stys.FunctionType.empty)
    args: list[stys.TypeArg] = Field(default_factory=list)

    def insert_port_types(self, in_types: TypeRow, out_types: TypeRow) -> None:
        self.signature = stys.FunctionType(input=list(in_types), output=list(out_types))

    def display_name(self) -> str:
        return self.name

    def deserialize(self) -> ops.Custom:
        return ops.Custom(
            extension=self.extension,
            op_name=self.name,
            signature=self.signature.deserialize(),
            args=deser_it(self.args),
        )

    model_config = ConfigDict(
        # Needed to avoid random '\n's in the pydantic description
        json_schema_extra={
            "description": (
                "A user-defined operation that can be downcasted by the extensions that"
                " define it."
            )
        }
    )


class Tag(DataflowOp):
    """An operation that creates a tagged sum value from one of its variants."""

    op: Literal["Tag"] = "Tag"
    tag: int  # The variant to create.
    variants: list[TypeRow]  # The variants of the sum type.

    def deserialize(self) -> ops.Tag:
        return ops.Tag(
            tag=self.tag,
            sum_ty=tys.Sum([deser_it(v) for v in self.variants]),
        )


class AliasDecl(BaseOp):
    op: Literal["AliasDecl"] = "AliasDecl"
    name: str
    bound: TypeBound

    def deserialize(self) -> ops.AliasDecl:
        return ops.AliasDecl(self.name, self.bound)


class AliasDefn(BaseOp):
    op: Literal["AliasDefn"] = "AliasDefn"
    name: str
    definition: Type

    def deserialize(self) -> ops.AliasDefn:
        return ops.AliasDefn(self.name, self.definition.deserialize())


class OpType(RootModel):
    """A constant operation."""

    root: (
        Module
        | Case
        | FuncDefn
        | FuncDecl
        | Const
        | DataflowBlock
        | ExitBlock
        | Conditional
        | TailLoop
        | CFG
        | Input
        | Output
        | Call
        | CallIndirect
        | LoadConstant
        | LoadFunction
        | ExtensionOp
        | Tag
        | DFG
        | AliasDecl
        | AliasDefn
    ) = Field(discriminator="op")

    model_config = ConfigDict(json_schema_extra={"required": ["parent", "op"]})


# Now that all classes are defined, we need to update the ForwardRefs in all type
# annotations. We use some inspect magic to find all classes defined in this file.
classes = (
    inspect.getmembers(
        sys.modules[__name__],
        lambda member: inspect.isclass(member) and member.__module__ == __name__,
    )
    + tys_classes
)

tys_model_rebuild(dict(classes))

from hugr import (  # noqa: E402 # needed to avoid circular imports
    ops,
    val,
)



"""Testing setup."""

import pytest

from hugr import ops, tys, val
from hugr.build import dfg
from hugr.hugr import base as hugr
from hugr.hugr import node_port


@pytest.fixture(autouse=True)
def _add_hugr(doctest_namespace):
    doctest_namespace.update(
        {
            "hugr": hugr,
            "node_port": node_port,
            "dfg": dfg,
            "ops": ops,
            "tys": tys,
            "val": val,
        }
    )



"""HUGR prelude values."""

from __future__ import annotations

from dataclasses import dataclass

from hugr import val
from hugr.std._util import _load_extension

PRELUDE_EXTENSION = _load_extension("prelude")

STRING_T_DEF = PRELUDE_EXTENSION.types["string"]

STRING_T = STRING_T_DEF.instantiate([])


@dataclass
class StringVal(val.ExtensionValue):
    """Custom value for a string."""

    v: str

    def to_value(self) -> val.Extension:
        name = "ConstString"
        return val.Extension(
            name,
            typ=STRING_T,
            val=self.v,
        )

    def __str__(self) -> str:
        return f"{self.v}"



from hugr import tys
from hugr.build.dfg import Dfg
from hugr.package import Package

from .conftest import QUANTUM_EXT, MeasureFree, QAlloc, validate


def test_order_links():
    dfg = Dfg(tys.Bool)
    inp_0 = dfg.input_node.out(0)
    inp_order = dfg.input_node.out(-1)
    out_0 = dfg.output_node.inp(0)
    out_1 = dfg.output_node.inp(1)
    out_order = dfg.output_node.inp(-1)

    dfg.hugr.add_link(inp_0, out_0)
    dfg.hugr.add_link(inp_0, out_1)
    assert list(dfg.hugr.outgoing_links(dfg.input_node)) == [
        (inp_0, [out_0, out_1]),
    ]
    assert list(dfg.hugr.incoming_links(dfg.output_node)) == [
        (out_0, [inp_0]),
        (out_1, [inp_0]),
    ]

    # Now add an order link
    dfg.hugr.add_order_link(dfg.input_node, dfg.output_node)
    assert list(dfg.hugr.incoming_order_links(dfg.output_node)) == [dfg.input_node]
    assert list(dfg.hugr.outgoing_order_links(dfg.input_node)) == [dfg.output_node]
    assert list(dfg.hugr.outgoing_links(dfg.input_node)) == [
        (inp_0, [out_0, out_1]),
        (inp_order, [out_order]),
    ]
    assert list(dfg.hugr.incoming_links(dfg.output_node)) == [
        (out_0, [inp_0]),
        (out_1, [inp_0]),
        (out_order, [inp_order]),
    ]


# https://github.com/CQCL/hugr/issues/2439
def test_order_unconnected(snapshot):
    dfg = Dfg(tys.Qubit)
    meas = dfg.add(MeasureFree(*dfg.inputs()))
    alloc = dfg.add_op(QAlloc)
    dfg.hugr.add_order_link(meas, alloc)
    dfg.set_outputs(alloc)

    validate(Package([dfg.hugr], [QUANTUM_EXT]), snap=snapshot)



"""Tests for the CLI bindings."""

from typing import Any

import pytest

from hugr import cli, tys
from hugr.build import Dfg, Module
from hugr.ext import Extension
from hugr.package import Package

from .serialization.test_extension import EXAMPLE


@pytest.fixture
def simple_hugr_bytes() -> bytes:
    """Create a simple HUGR package as bytes for testing."""
    return Package([Module().hugr]).to_bytes()


@pytest.fixture
def hugr_with_extension_bytes() -> bytes:
    """Create a HUGR package with an extension as bytes for testing."""
    ext = Extension.from_json(EXAMPLE)
    module = Module()
    return Package([module.hugr], [ext]).to_bytes()


def test_validate_with_bytes(simple_hugr_bytes: bytes):
    """Test validating a HUGR package using the programmatic API."""
    cli.validate(simple_hugr_bytes)


def test_validate_with_bytes_invalid():
    """Test that invalid packages raise errors through the programmatic API."""
    # We need to pass invalid bytes through cli_with_io directly
    # since Package construction would fail first

    invalid_bytes = b"not a valid hugr package"

    with pytest.raises(cli.HugrCliError, match="Bad magic number"):
        cli.cli_with_io(["validate"], invalid_bytes)


def test_validate_no_std(simple_hugr_bytes: bytes):
    """Test validate with no_std flag."""
    cli.validate(simple_hugr_bytes, no_std=True)


@pytest.mark.parametrize(
    "kwargs",
    [
        {"format": "json"},  # convert to JSON format
        {"text": True},  # convert to text format
        {"compress": True, "compression_level": 9},  # convert with compression
    ],
)
def test_convert_format(hugr_with_extension_bytes: bytes, kwargs: dict[str, Any]):
    """Test converting a HUGR package between formats."""
    output_bytes = cli.convert(hugr_with_extension_bytes, **kwargs)

    output_package = Package.from_bytes(output_bytes)
    input_package = Package.from_bytes(hugr_with_extension_bytes)
    assert output_package == input_package


def test_mermaid_output(simple_hugr_bytes: bytes):
    """Test generating mermaid diagrams from a HUGR package."""
    output = cli.mermaid(simple_hugr_bytes)

    assert "graph LR" in output


def test_mermaid_with_validation(simple_hugr_bytes: bytes):
    """Test generating mermaid diagrams with validation."""
    output = cli.mermaid(simple_hugr_bytes, validate=True)
    assert "graph LR" in output


def test_describe_output(simple_hugr_bytes: bytes):
    """Test describing a HUGR package."""
    output_text = cli.describe_str(simple_hugr_bytes)

    # Should contain package information
    assert "Package contains" in output_text


def test_describe_with_options(hugr_with_extension_bytes: bytes):
    """Test describe with various options."""
    output_text = cli.describe_str(hugr_with_extension_bytes, packaged_extensions=True)
    assert "Packaged extensions:" in output_text

    output_text = cli.describe_str(
        hugr_with_extension_bytes, no_resolved_extensions=True
    )
    assert "resolved" not in output_text

    output_text = cli.describe_str(hugr_with_extension_bytes, public_symbols=True)
    assert len(output_text) > 0

    # Test with generator_claimed_extensions flag
    output_text = cli.describe_str(
        hugr_with_extension_bytes, generator_claimed_extensions=True
    )
    assert len(output_text) > 0


def test_describe_json_basic(simple_hugr_bytes: bytes):
    """Test describe_json returns structured PackageDesc."""
    desc = cli.describe(simple_hugr_bytes)

    assert isinstance(desc, cli.PackageDesc)

    # Should have expected fields
    assert desc.header is not None
    assert isinstance(desc.modules, list)
    assert len(desc.modules) == 1

    # Module should have properties
    module = desc.modules[0]
    assert module is not None
    assert isinstance(module, cli.ModuleDesc)
    assert module.num_nodes is not None
    assert module.num_nodes > 0


def test_describe_json_with_packaged_extensions(hugr_with_extension_bytes: bytes):
    """Test describe_json with packaged_extensions flag."""
    desc = cli.describe(hugr_with_extension_bytes, packaged_extensions=True)

    # Should have packaged_extensions field populated
    assert isinstance(desc, cli.PackageDesc)
    assert desc.packaged_extensions is not None

    mod = desc.modules[0]
    assert mod is not None
    # mock use of extension in module
    mod.used_extensions_resolved = desc.packaged_extensions  # type: ignore[assignment]

    assert desc.uses_extension("ext")
    assert not desc.uses_extension("nonexistent_extension")


@pytest.fixture
def hugr_using_ext() -> bytes:
    """A simple HUGR package that uses an extension, but doesn't package it."""
    ext = Extension.from_json(EXAMPLE)
    u_t = tys.USize()
    op = ext.get_op("New").instantiate(
        [u_t.type_arg()], concrete_signature=tys.FunctionType([u_t], [])
    )
    h = Dfg(u_t)
    a = h.inputs()[0]
    h.add_op(op, a)
    h.set_outputs()

    package = Package([h.hugr], [])

    return package.to_bytes()


def test_failed_describe(hugr_using_ext):
    """Json description still succeeds, with error field populated"""
    desc = cli.describe(hugr_using_ext)
    mod = desc.modules[0]
    assert mod is not None
    assert mod.num_nodes == 8  # computed before error
    assert isinstance(desc.error, str)
    assert "requires extension ext" in desc.error

    assert desc.uses_extension("ext")



"""Types and operations from the standard extension set."""

from hugr.ext import Extension, ExtensionRegistry

from . import collections, float, int, logic, prelude, ptr

__all__ = ["PRELUDE", "collections", "float", "int", "logic", "prelude", "ptr"]

PRELUDE: Extension = prelude.PRELUDE_EXTENSION


def _std_extensions() -> ExtensionRegistry:
    return ExtensionRegistry.from_extensions(
        [
            PRELUDE,
            int.CONVERSIONS_EXTENSION,
            int.INT_TYPES_EXTENSION,
            int.INT_OPS_EXTENSION,
            float.FLOAT_OPS_EXTENSION,
            float.FLOAT_TYPES_EXTENSION,
            logic.EXTENSION,
            ptr.EXTENSION,
            collections.array.EXTENSION,
            collections.borrow_array.EXTENSION,
            collections.list.EXTENSION,
            collections.static_array.EXTENSION,
        ]
    )



"""Generic serialization of HUGRs and Packages.

The format is designed to be extensible and backwards-compatible. It
consists of a header declaring the format used to encode the HUGR, followed
by the encoded HUGR itself.

Payload formats
---------------

The envelope may encode the HUGR in different formats, listed in
:class:`hugr.envelope.EnvelopeFormat`. The payload may also be compressed with zstd.

Some formats can be represented as ASCII, as indicated by the
:meth:`hugr.envelope.EnvelopeFormat.ascii_printable` method. When this is the case, the
whole envelope can be stored in a string.

Envelope header
---------------

The binary header format is 10 bytes, with the following fields:

+--------+--------------+--------------------------------------------------------------------------------+
| Field  | Size (bytes) | Description                                                                    |
+========+==============+================================================================================+
| Magic  | 8            | :class:`hugr.envelope.MAGIC_NUMBERS` constant identifying the envelope format. |
| Format | 1            | :class:`hugr.envelope.EnvelopeFormat` describing the payload format.           |
| Flags  | 1            | Additional configuration flags.                                                |
+--------+--------------+--------------------------------------------------------------------------------+

Flags:

- Bit 0: Whether the payload is compressed with zstd.
- Bits 1-5: Reserved for future use.
- Bits 7,6: Constant "01" to make some headers ascii-printable.
"""  # noqa: E501

from __future__ import annotations

import json
from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, Any, ClassVar

from semver import Version
from typing_extensions import deprecated

import hugr._hugr.model as rust
import hugr._hugr.zstd as zstd

if TYPE_CHECKING:
    from hugr.hugr.base import Hugr
    from hugr.package import Package

# This is a hard-coded magic number that identifies the start of a HUGR envelope.
MAGIC_NUMBERS = b"HUGRiHJv"

# The all-unset header flags configuration.
# Bit 7 is always set to ensure we have a printable ASCII character.
_DEFAULT_FLAGS = 0b0100_0000
# The ZSTD flag bit in the header's flags.
_ZSTD_FLAG = 0b0000_0001


@deprecated("Use Package/Hugr.to_bytes() instead.")
def make_envelope(package: Package | Hugr, config: EnvelopeConfig) -> bytes:
    """Encode a Package into an envelope, using the given configuration."""
    from hugr.hugr.base import Hugr

    if isinstance(package, Hugr):
        package = package.to_package()
    return _make_envelope(package, config)


def _make_envelope(package: Package, config: EnvelopeConfig) -> bytes:
    """Encode a Package into an envelope, using the given configuration."""
    envelope = bytearray(config._make_header().to_bytes())

    payload: bytes
    match config.format:
        case EnvelopeFormat.JSON:
            json_str = package._to_serial().model_dump_json()
            # This introduces an extra encode/decode roundtrip when calling
            # `make_envelope_str`, but we prioritize speed for binary formats.
            payload = json_str.encode("utf-8")

        case EnvelopeFormat.MODEL:
            payload = bytes(package.to_model())

        case EnvelopeFormat._S_EXPRESSION:
            payload = str(package.to_model()).encode("utf-8")

        case EnvelopeFormat.MODEL_WITH_EXTS:
            # Encoded model first, followed by json-encoded extensions.
            package_bytes = bytes(package.to_model())
            extension_str = json.dumps(
                [ext._to_serial().model_dump(mode="json") for ext in package.extensions]
            )
            extension_bytes = extension_str.encode("utf8")
            payload = package_bytes + extension_bytes

        case EnvelopeFormat._S_EXPRESSION_WITH_EXTS:
            # Json-encoded extensions first, followed by s-expression.
            # (Due to restrictions in S-expression parsing)
            extension_str = json.dumps(
                [ext._to_serial().model_dump(mode="json") for ext in package.extensions]
            )
            package_str = str(package.to_model())
            payload_str = extension_str + package_str
            payload = payload_str.encode("utf-8")

    if config.zstd is not None:
        payload = zstd.compress(payload, config.zstd)

    envelope += payload
    return bytes(envelope)


@deprecated("Use Package/Hugr.to_str() instead.")
def make_envelope_str(package: Package | Hugr, config: EnvelopeConfig) -> str:
    """Encode a Package into an envelope, using the given configuration."""
    from hugr.hugr.base import Hugr

    if isinstance(package, Hugr):
        package = package.to_package()
    return _make_envelope_str(package, config)


def _make_envelope_str(package: Package, config: EnvelopeConfig) -> str:
    """Encode a Package into an envelope, using the given configuration."""
    if not config.format.ascii_printable():
        msg = "Only ascii-printable envelope formats can be encoded into a string."
        raise ValueError(msg)
    envelope = _make_envelope(package, config)
    return envelope.decode("utf-8")


def read_envelope(envelope: bytes) -> Package:
    """Decode a HUGR package from an envelope."""
    import hugr._serialization.extension as ext_s
    from hugr.package import Package

    header = EnvelopeHeader.from_bytes(envelope)
    payload = envelope[10:]

    if header.zstd:
        payload = zstd.decompress(payload)

    match header.format:
        case EnvelopeFormat.JSON:
            return ext_s.Package.model_validate_json(payload).deserialize()
        case EnvelopeFormat.MODEL | EnvelopeFormat._S_EXPRESSION:
            model_package, suffix = rust.bytes_to_package(payload)
            if suffix:
                msg = f"Excess bytes in envelope with format {EnvelopeFormat.MODEL}."
                raise ValueError(msg)
            return Package.from_model(model_package)
        case EnvelopeFormat.MODEL_WITH_EXTS:
            # Encoded model first, followed by json-encoded extensions.
            from hugr.ext import Extension

            model_package, suffix = rust.bytes_to_package(payload)
            return Package(
                modules=Package.from_model(model_package).modules,
                extensions=[
                    Extension.from_json(json.dumps(extension))
                    for extension in json.loads(suffix)
                ],
            )
        case EnvelopeFormat._S_EXPRESSION_WITH_EXTS:
            # Json-encoded extensions first, followed by s-expression.
            # (Due to restrictions in S-expression parsing)
            from hugr.ext import Extension

            decoder = json.JSONDecoder()
            payload_str = payload.decode("utf-8")
            extensions, parse_index = decoder.raw_decode(payload_str)

            extensions = [
                Extension.from_json(json.dumps(extension)) for extension in extensions
            ]

            model_package = rust.string_to_package(payload_str[parse_index:])

            return Package(
                modules=Package.from_model(model_package).modules,
                extensions=extensions,
            )


def read_envelope_hugr(envelope: bytes) -> Hugr:
    """Decode a HUGR from an envelope.

    Raises:
        ValueError: If the envelope does not contain a single module.
    """
    pkg = read_envelope(envelope)
    if len(pkg.modules) != 1:
        msg = (
            "Expected a single module in the envelope, but got "
            + f"{len(pkg.modules)} modules."
        )
        raise ValueError(msg)
    return pkg.modules[0]


def read_envelope_str(envelope: str) -> Package:
    """Decode a HUGR package from an envelope."""
    return read_envelope(envelope.encode("utf-8"))


def read_envelope_hugr_str(envelope: str) -> Hugr:
    """Decode a HUGR from an envelope.

    Raises:
        ValueError: If the envelope does not contain a single module.
    """
    pkg = read_envelope_str(envelope)
    if len(pkg.modules) != 1:
        msg = (
            "Expected a single module in the envelope, but got "
            + f"{len(pkg.modules)} modules."
        )
        raise ValueError(msg)
    return pkg.modules[0]


class EnvelopeFormat(Enum):
    """Format used to encode a HUGR envelope."""

    MODEL = 1
    """A capnp-encoded hugr-model."""
    MODEL_WITH_EXTS = 2
    """A capnp-encoded hugr-model, immediately followed by a json-encoded
    extension registry."""
    _S_EXPRESSION = 40
    """A textual representation of a hugr.

    Note: This format is **experimental** and not fully tested.
    It is not recommended for production use.
    """
    _S_EXPRESSION_WITH_EXTS = 41
    """A textual representation of a hugr, preceded by a json-encoded
    extension registry.

    Note: This format is **experimental** and not fully tested.
    It is not recommended for production use.
    """
    JSON = 63  # '?' in ASCII
    """A json-encoded hugr-package. This format is ASCII-printable."""

    def ascii_printable(self) -> bool:
        return self in {
            EnvelopeFormat.JSON,
            EnvelopeFormat._S_EXPRESSION,
            EnvelopeFormat._S_EXPRESSION_WITH_EXTS,
        }


@dataclass
class EnvelopeHeader:
    """Header of a HUGR envelope.

    See the module docstring for the binary format.

    Attributes:
        format: The format used to encode the HUGR.
        zstd: Whether the payload is compressed with zstd.
            The compression level is detected from the payload.
    """

    format: EnvelopeFormat
    zstd: bool = False

    def to_bytes(self) -> bytes:
        header_bytes = bytearray(MAGIC_NUMBERS)
        header_bytes.append(self.format.value)
        flags = _DEFAULT_FLAGS
        if self.zstd:
            flags |= _ZSTD_FLAG
        header_bytes.append(flags)
        return bytes(header_bytes)

    @staticmethod
    def from_bytes(data: bytes) -> EnvelopeHeader:
        if len(data) < 10:
            msg = (
                "Invalid HUGR envelope."
                + f"Expected at least 10 bytes, but got {len(data)} bytes."
            )
            raise ValueError(msg)
        if data[:8] != MAGIC_NUMBERS:
            msg = (
                "Invalid magic number for HUGR envelope."
                + f"Expected {MAGIC_NUMBERS!r} but got {data[:8]!r}"
            )
            raise ValueError(msg)

        format: EnvelopeFormat = EnvelopeFormat(data[8])

        flags = data[9]
        zstd = bool(flags & _ZSTD_FLAG)
        other_flags = (flags ^ _DEFAULT_FLAGS) & ~_ZSTD_FLAG
        if other_flags:
            flag_ids = [i for i in range(8) if other_flags & (1 << i)]
            msg = (
                f"Unrecognised Envelope flags {flag_ids}."
                + " Please update your HUGR version."
            )
            raise ValueError(msg)

        return EnvelopeHeader(format=format, zstd=zstd)


@dataclass
class EnvelopeConfig:
    """Configuration for writing a HUGR envelope.

    Attributes:
        format: The format to use for encoding the HUGR.
        zstd: The compression level to use, or None for no compression.
            Use 0 for the default level.
    """

    format: EnvelopeFormat = EnvelopeFormat.JSON
    zstd: int | None = None

    TEXT: ClassVar[EnvelopeConfig]
    BINARY: ClassVar[EnvelopeConfig]

    def _make_header(self) -> EnvelopeHeader:
        return EnvelopeHeader(format=self.format, zstd=self.zstd is not None)


# Set EnvelopeConfig's class variables.
# These can only be initialized _after_ the class is defined.
EnvelopeConfig.TEXT = EnvelopeConfig(format=EnvelopeFormat.JSON, zstd=None)
EnvelopeConfig.BINARY = EnvelopeConfig(format=EnvelopeFormat.MODEL_WITH_EXTS, zstd=0)


@dataclass(frozen=True)
class GeneratorDesc:
    """Description of the generator that defined the HUGR module.

    These are stored at the module root node metadata under the
    :class:`hugr.metadata.HugrGenerator` entry.
    """

    name: str
    version: Version | None

    def to_json(self) -> dict[str, str]:
        """Encodes the generator as a dictionary of native types that can be
        serialized by `json.dump`.
        """
        if self.version is None:
            return {
                "name": self.name,
            }
        else:
            return {
                "name": self.name,
                "version": str(self.version),
            }

    @classmethod
    def from_json(cls, value: Any) -> GeneratorDesc:
        """Decodes the generator from a native types obtained from `json.load`."""
        if isinstance(value, str):
            return GeneratorDesc(name=value, version=None)

        if not isinstance(value, dict):
            msg = (
                "Expected generator metadata to be a string or a dict,"
                + " but got {type(value)}"
            )
            raise TypeError(msg)

        fallback_name = " ".join(f"{k}: {v}" for k, v in value.items())
        if "name" not in value or any(k != "name" and k != "version" for k in value):
            return GeneratorDesc(name=fallback_name, version=None)
        if "version" in value:
            try:
                version = Version.parse(value["version"])
            except ValueError:
                return GeneratorDesc(name=fallback_name, version=None)
            return GeneratorDesc(name=value["name"], version=version)
        else:
            return GeneratorDesc(name=value["name"], version=None)


@dataclass
class ExtensionDesc:
    """High level description of a HUGR extension.

    A list of these is stored at the module root node metadata under the
    :class:`hugr.metadata.HugrUsedExtensions` entry.
    """

    name: str
    version: Version

    def to_json(self) -> dict[str, str]:
        """Encodes the extension as a dictionary of native types that can be
        serialized by `json.dump`.
        """
        return {
            "name": self.name,
            "version": str(self.version),
        }

    @classmethod
    def from_json(cls, value: Any) -> ExtensionDesc:
        """Decodes the extension from a native types obtained from `json.load`."""
        if not isinstance(value, dict):
            msg = f"Expected extension metadata to be a dict, but got {type(value)}"
            raise TypeError(msg)
        if "name" not in value:
            msg = (
                "Expected extension metadata to be a dict with a 'name' key,"
                + f" but got {value}"
            )
            raise TypeError(msg)
        if "version" not in value:
            msg = (
                "Expected extension metadata to be a dict with a 'version' key,"
                + f" but got {value}"
            )
            raise TypeError(msg)
        return ExtensionDesc(
            name=value["name"], version=Version.parse(value["version"])
        )



"""Statically sized immutable array type and its operations."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from hugr import tys, val
from hugr.std._util import _load_extension
from hugr.utils import comma_sep_str

if TYPE_CHECKING:
    from hugr.ext import ExtensionRegistry, ExtensionResolutionResult

EXTENSION = _load_extension("collections.static_array")


@dataclass(eq=False)
class StaticArray(tys.ExtType):
    """Fixed size immutable array of `ty` elements."""

    def __init__(self, ty: tys.Type) -> None:
        self.type_def = EXTENSION.types["static_array"]
        if (
            tys.TypeBound.join(ty.type_bound(), tys.TypeBound.Copyable)
            != tys.TypeBound.Copyable
        ):
            msg = "Static array elements must be copyable"
            raise ValueError(msg)
        self.args = [tys.TypeTypeArg(ty)]

    @property
    def ty(self) -> tys.Type:
        assert isinstance(
            self.args[0], tys.TypeTypeArg
        ), "Array elements must have a valid type"
        return self.args[0].ty

    def type_bound(self) -> tys.TypeBound:
        return self.ty.type_bound()

    def _resolve_used_extensions(
        self, registry: ExtensionRegistry | None = None
    ) -> tuple[StaticArray, ExtensionResolutionResult]:
        ext_type, result = super()._resolve_used_extensions(registry)

        assert isinstance(
            ext_type, tys.ExtType
        ), "HUGR internal error, expected resolved type to be extension type."
        assert (
            ext_type.type_def == EXTENSION.types["static_array"]
        ), "HUGR internal error, expected resolved type to be static array."

        static_array = StaticArray(tys.Unit)
        static_array.args = ext_type.args
        return static_array, result


@dataclass
class StaticArrayVal(val.ExtensionValue):
    """Constant value for a statically sized immutable array of elements."""

    v: list[val.Value]
    ty: StaticArray
    name: str

    def __init__(self, v: list[val.Value], elem_ty: tys.Type, name: str) -> None:
        self.v = v
        self.ty = StaticArray(elem_ty)
        self.name = name

    def to_value(self) -> val.Extension:
        # Encode the nested values as JSON strings directly, to mirror what
        # happens when loading (where we can't decode the constant payload back
        # into specialized `Value`s).
        serial_val = {
            "value": {
                "values": [v._to_serial_root() for v in self.v],
                "typ": self.ty.ty._to_serial_root(),
            },
            "name": self.name,
        }
        return val.Extension("StaticArrayValue", typ=self.ty, val=serial_val)

    def __str__(self) -> str:
        return f"static_array({comma_sep_str(self.v)})"

    def _resolve_used_extensions_inplace(
        self, registry: ExtensionRegistry | None = None
    ) -> ExtensionResolutionResult:
        resolved_ty, result = self.ty._resolve_used_extensions(registry)
        assert isinstance(
            resolved_ty, StaticArray
        ), "HUGR internal error, expected resolved type to be static array."
        self.ty = resolved_ty
        for value in self.v:
            result.extend(value._resolve_used_extensions_inplace(registry))
        return result



"""Python interface for the HUGR CLI subcommands.

Provides programmatic access to the HUGR CLI via Rust bindings.
Exposes a generic `cli_with_io` function and helpers for the main subcommands:
validate, describe, convert, and mermaid.
"""

from pydantic import BaseModel

from hugr._hugr.model import HugrCliDescribeError, HugrCliError, cli_with_io

__all__ = [
    "cli_with_io",
    "validate",
    "describe_str",
    "describe",
    "convert",
    "mermaid",
    "PackageDesc",
    "ModuleDesc",
    "ExtensionDesc",
    "EntrypointDesc",
    "HugrCliError",
    "HugrCliDescribeError",
]


def _add_input_args(
    args: list[str], no_std: bool, extensions: list[str] | None
) -> list[str]:
    """Add common HugrInputArgs parameters to the argument list."""
    if no_std:
        args.append("--no-std")
    if extensions is not None:
        for ext in extensions:
            args.extend(["--extensions", ext])
    return args


def validate(
    hugr_bytes: bytes,
    *,
    no_std: bool = False,
    extensions: list[str] | None = None,
) -> None:
    """Validate a HUGR package.

    Args:
        hugr_bytes: The HUGR package as bytes.
        no_std: Don't use standard extensions when validating hugrs.
            Prelude is still used (default: False).
        extensions: Paths to additional serialised extensions needed to load the HUGR.

    Raises:
        HugrCliError: On validation failure or other CLI errors.
    """
    args = _add_input_args(["validate"], no_std, extensions)
    cli_with_io(args, hugr_bytes)


class EntrypointDesc(BaseModel):
    """Description of a module's entrypoint node.

    Attributes:
        node: The node index of the entrypoint.
        optype: String representation of the operation type.
    """

    node: int
    optype: str


class ExtensionDesc(BaseModel):
    """Description of a HUGR extension.

    Attributes:
        name: The name of the extension.
        version: The version string of the extension.
    """

    name: str
    version: str


class ModuleDesc(BaseModel):
    """Description of a HUGR module.

    Attributes:
        entrypoint: The entrypoint node of the module, if present.
        generator: Name and version of the generator that created this module.
        num_nodes: Total number of nodes in the module.
        public_symbols: List of public symbol names exported by the module.
        used_extensions_generator: Extensions claimed by the generator in metadata.
        used_extensions_resolved: Extensions actually used by the module operations.
    """

    entrypoint: EntrypointDesc | None = None
    generator: str | None = None
    num_nodes: int | None = None
    public_symbols: list[str] | None = None
    used_extensions_generator: list[ExtensionDesc] | None = None
    used_extensions_resolved: list[ExtensionDesc] | None = None

    def uses_extension(self, extension_name: str) -> bool:
        """Check if this module uses a specific extension.

        Args:
            extension_name: The name of the extension to check.

        Returns:
            True if the module uses the extension, False otherwise.
        """
        return any(
            ext.name == extension_name for ext in self.used_extensions_resolved or []
        )


class PackageDesc(BaseModel):
    """Description of a HUGR package.

    Attributes:
        error: Error message if the package failed to load.
        header: String representation of the envelope header.
        modules: List of module descriptions in the package.
        packaged_extensions: Extensions bundled with the package.
    """

    error: str | None = None
    header: str
    modules: list[ModuleDesc | None]
    packaged_extensions: list[ExtensionDesc | None] | None = None

    def uses_extension(self, extension_name: str) -> bool:
        """Check if any module in this package uses a specific extension.

        Args:
            extension_name: The name of the extension to check.

        Returns:
            True if any module uses the extension, False otherwise.
        """
        return any(
            module.uses_extension(extension_name)
            for module in self.modules
            if module is not None
        )


def describe_str(
    hugr_bytes: bytes,
    *,
    packaged_extensions: bool = False,
    no_resolved_extensions: bool = False,
    public_symbols: bool = False,
    generator_claimed_extensions: bool = False,
    no_std: bool = False,
    extensions: list[str] | None = None,
    _json: bool = False,  # only used by describe()
) -> str:
    """Describe the contents of a HUGR package as text.

    If an error occurs during loading, partial descriptions are printed.
    For example, if the first module is loaded and the second fails,
    then only the first module will be described.

    Args:
        hugr_bytes: The HUGR package as bytes.
        packaged_extensions: Enumerate packaged extensions (default: False).
        no_resolved_extensions: Don't display resolved extensions used by the module
            (default: False).
        public_symbols: Display public symbols in the module (default: False).
        generator_claimed_extensions: Display claimed extensions set by generator
            in module metadata (default: False).
        no_std: Don't use standard extensions when validating hugrs.
            Prelude is still used (default: False).
        extensions: Paths to additional serialised extensions needed to load the HUGR.

    Returns:
        Text description of the package.

    Raises:
        HugrCliDescribeError: On error during package description. The exception
            contains partial output if available.
    """
    args = ["describe"]
    if _json:
        args.append("--json")
    if packaged_extensions:
        args.append("--packaged-extensions")
    if no_resolved_extensions:
        args.append("--no-resolved-extensions")
    if public_symbols:
        args.append("--public-symbols")
    if generator_claimed_extensions:
        args.append("--generator-claimed-extensions")
    args = _add_input_args(args, no_std, extensions)
    return cli_with_io(args, hugr_bytes).decode("utf-8")


def describe(
    hugr_bytes: bytes,
    *,
    packaged_extensions: bool = False,
    no_resolved_extensions: bool = False,
    public_symbols: bool = False,
    generator_claimed_extensions: bool = False,
    no_std: bool = False,
    extensions: list[str] | None = None,
) -> PackageDesc:
    """Describe the contents of a HUGR package.

    If an error occurs during loading, partial descriptions are returned.
    For example, if the first module is loaded and the second fails,
    then only the first module will be described.

    Args:
        hugr_bytes: The HUGR package as bytes.
        packaged_extensions: Enumerate packaged extensions (default: False).
        no_resolved_extensions: Don't display resolved extensions used by the module
            (default: False).
        public_symbols: Display public symbols in the module (default: False).
        generator_claimed_extensions: Display claimed extensions set by generator
            in module metadata (default: False).
        no_std: Don't use standard extensions when validating hugrs.
            Prelude is still used (default: False).
        extensions: Paths to additional serialised extensions needed to load the HUGR.

    Returns:
        Structured package description as a PackageDesc object.
    """
    try:
        output = describe_str(
            hugr_bytes,
            _json=True,
            packaged_extensions=packaged_extensions,
            no_resolved_extensions=no_resolved_extensions,
            public_symbols=public_symbols,
            generator_claimed_extensions=generator_claimed_extensions,
            no_std=no_std,
            extensions=extensions,
        )
    except HugrCliDescribeError as e:
        output = e.args[1]
    return PackageDesc.model_validate_json(output)


def convert(
    hugr_bytes: bytes,
    *,
    format: str | None = None,
    text: bool = False,
    binary: bool = False,
    compress: bool = False,
    compression_level: int | None = None,
    no_std: bool = False,
    extensions: list[str] | None = None,
) -> bytes:
    """Convert between different HUGR envelope formats.

    Args:
        hugr_bytes: The HUGR package as bytes.
        format: Output format. One of: json, model, model-exts, s-expression,
            s-expression-exts (default: None, meaning same format as input).
        text: Use default text-based envelope configuration. Cannot be combined
            with format or binary (default: False).
        binary: Use default binary envelope configuration. Cannot be combined
            with format or text (default: False).
        compress: Enable zstd compression for the output (default: False).
        compression_level: Zstd compression level (1-22, where 1 is fastest and
            22 is best compression). (default None, uses the zstd default).
        no_std: Don't use standard extensions when validating hugrs.
            Prelude is still used (default: False).
        extensions: Paths to additional serialised extensions needed to load the HUGR.

    Returns:
        Converted package as bytes.

    Raises:
        HugrCliError: On conversion failure or other CLI errors.
    """
    args = ["convert"]
    if format is not None:
        args.extend(["--format", format])
    if text:
        args.append("--text")
    if binary:
        args.append("--binary")
    if compress:
        args.append("--compress")
    if compression_level is not None:
        args.extend(["--compression-level", str(compression_level)])
    args = _add_input_args(args, no_std, extensions)
    return cli_with_io(args, hugr_bytes)


def mermaid(
    hugr_bytes: bytes,
    *,
    validate: bool = False,
    no_std: bool = False,
    extensions: list[str] | None = None,
) -> str:
    """Generate mermaid diagrams from a HUGR package.

    Args:
        hugr_bytes: The HUGR package as bytes.
        validate: Validate before rendering, includes extension inference
            (default: False).
        no_std: Don't use standard extensions when validating hugrs.
            Prelude is still used (default: False).
        extensions: Paths to additional serialised extensions needed to load the HUGR.

    Returns:
        Mermaid diagram output as a string.

    Raises:
        HugrCliError: On mermaid generation failure or other CLI errors.
    """
    args = ["mermaid"]
    if validate:
        args.append("--validate")
    args = _add_input_args(args, no_std, extensions)
    return cli_with_io(args, hugr_bytes).decode("utf-8")



"""HUGR logic operations."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, ClassVar

from hugr.ops import Command, DataflowOp, RegisteredOp
from hugr.std._util import _load_extension

if TYPE_CHECKING:
    from hugr import ext
    from hugr.ops import ComWire


EXTENSION = _load_extension("logic")


@dataclass(frozen=True)
class _NotOp(RegisteredOp):
    """Logical NOT operation."""

    const_op_def: ClassVar[ext.OpDef] = EXTENSION.operations["Not"]

    def __call__(self, a: ComWire) -> Command:
        return DataflowOp.__call__(self, a)


#: Not operation
Not = _NotOp()



import pytest

from hugr.build.dfg import Dfg
from hugr.std.int import IntVal, int_t
from hugr.std.prelude import STRING_T, StringVal

from .conftest import validate


def test_string_val():
    val = StringVal("test")
    ext_val = val.to_value()
    assert ext_val.name == "ConstString"
    assert ext_val.typ == STRING_T
    assert ext_val.val == "test"

    dfg = Dfg()
    v = dfg.load(val)
    dfg.set_outputs(v)

    validate(dfg.hugr)


@pytest.mark.parametrize(
    ("log_width", "v", "unsigned"),
    [
        (5, 1, 1),
        (4, 0, 0),
        (6, 42, 42),
        (2, -1, 15),
        (1, -2, 2),
        (3, -23, 233),
        (3, -256, None),
        (2, 16, None),
    ],
)
def test_int_val(log_width: int, v: int, unsigned: int | None):
    val = IntVal(v, log_width)
    if unsigned is None:
        with pytest.raises(
            ValueError,
            match=f"Value {v} out of range for {1<<log_width}-bit signed integer.",
        ):
            val.to_value()
        return
    ext_val = val.to_value()

    assert ext_val.name == "ConstInt"
    assert ext_val.typ == int_t(log_width)
    assert ext_val.val == {"log_width": log_width, "value": unsigned}

    dfg = Dfg()
    o = dfg.load(val)
    dfg.set_outputs(o)

    validate(dfg.hugr)



from hugr import tys
from hugr.build import Dfg
from hugr.build.function import Module
from hugr.ext import Extension, FixedHugr, OpDef, OpDefSig, Version
from hugr.package import (
    FuncDeclPointer,
    FuncDefnPointer,
    ModulePointer,
    Package,
    PackagePointer,
)

from .conftest import validate


def test_package():
    mod = Module()
    f_id = mod.define_function("id", [tys.Qubit])
    f_id.set_outputs(f_id.input_node[0])

    mod2 = Module()
    f_id_decl = mod2.declare_function(
        "id", tys.PolyFuncType([], tys.FunctionType([tys.Qubit], [tys.Qubit]))
    )
    f_main = mod2.define_main([tys.Qubit])
    q = f_main.input_node[0]
    call = f_main.call(f_id_decl, q)
    f_main.set_outputs(call)

    package = Package([mod.hugr, mod2.hugr])
    validate(package)

    p = PackagePointer(package)
    assert p.package == package

    m = ModulePointer(package, 1)
    assert m.module == mod2.hugr

    f = FuncDeclPointer(package, 1, f_id_decl)
    assert f.func_decl == mod2.hugr[f_id_decl].op

    f = FuncDefnPointer(package, 0, f_id.to_node())

    assert f.func_defn == mod.hugr[f_id.to_node()].op

    main = m.to_executable_package()
    assert main.entry_point_node == f_main.to_node()


def test_lower_func():
    hugr = Dfg(tys.Qubit)
    hugr.set_outputs(hugr.input_node[0])

    ext = Extension("dummy", Version(0, 1, 0))
    ext.add_op_def(
        OpDef(
            "dummy_op",
            OpDefSig(tys.FunctionType.endo([tys.Qubit])),
            lower_funcs=[FixedHugr([], hugr.hugr)],
        )
    )

    pkg = Package([hugr.hugr], [ext])

    validate(pkg)



from __future__ import annotations

import os
from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, TypeVar

from typing_extensions import Self

from hugr import cli, ext, ops, tys
from hugr.envelope import EnvelopeConfig, EnvelopeFormat
from hugr.hugr import Hugr
from hugr.ops import AsExtOp, Command, Const, Custom, DataflowOp, ExtOp, RegisteredOp
from hugr.package import Package
from hugr.std.float import FLOAT_T

if TYPE_CHECKING:
    import typing

    from syrupy.assertion import SnapshotAssertion

    from hugr.hugr.node_port import Node
    from hugr.ops import ComWire

QUANTUM_EXT = ext.Extension("pytest.quantum", ext.Version(0, 1, 0))
QUANTUM_EXT.add_op_def(
    ext.OpDef(
        name="H",
        description="Hadamard gate",
        signature=ext.OpDefSig(tys.FunctionType.endo([tys.Qubit])),
    )
)

QUANTUM_EXT.add_op_def(
    ext.OpDef(
        name="CX",
        description="CNOT gate",
        signature=ext.OpDefSig(tys.FunctionType.endo([tys.Qubit] * 2)),
    )
)

# Shared test extension
TEST_EXT = ext.Extension("pytest.test_ext", ext.Version(0, 1, 0))


def test_ext_registry() -> ext.ExtensionRegistry:
    """Registry containing TEST_EXT."""
    registry = ext.ExtensionRegistry()
    registry.add_extension(TEST_EXT)
    return registry


TEST_TYPE_DEF = TEST_EXT.add_type_def(
    ext.TypeDef(
        name="TestType",
        description="A custom type for testing",
        params=[],
        bound=ext.ExplicitBound(tys.TypeBound.Copyable),
    )
)
TEST_OP_DEF = TEST_EXT.add_op_def(
    ext.OpDef(
        name="TestOp",
        description="A custom operation for testing",
        signature=ext.OpDefSig(tys.FunctionType([tys.Bool], [tys.Bool])),
    )
)
TEST_OP = ops.ExtOp(TEST_OP_DEF, args=[])
TEST_TYPE = tys.ExtType(TEST_TYPE_DEF, args=[])

# Unresolved opaque type and Custom op for testing resolution
TEST_TYPE_OPAQUE = tys.Opaque(
    id="TestType", bound=tys.TypeBound.Copyable, extension=TEST_EXT.name
)
TEST_OP_OPAQUE = ops.Custom(
    op_name="TestOp",
    signature=tys.FunctionType([tys.Bool], [tys.Bool]),
    extension=TEST_EXT.name,
)


E = TypeVar("E", bound=Enum)


def _load_enum(enum_cls: type[E], custom: ExtOp) -> E | None:
    ext = custom._op_def._extension
    assert ext is not None
    if ext.name == QUANTUM_EXT.name and custom._op_def.name in enum_cls.__members__:
        return enum_cls(custom._op_def.name)
    return None


@dataclass(frozen=True)
class OneQbGate(AsExtOp):
    # Have to nest enum to avoid meta class conflict
    class _Enum(Enum):
        H = "H"

    _enum: _Enum

    def __call__(self, q: ComWire) -> Command:
        return DataflowOp.__call__(self, q)

    def op_def(self) -> ext.OpDef:
        return QUANTUM_EXT.operations[self._enum.value]

    @classmethod
    def from_ext(cls, custom: ExtOp) -> Self | None:
        return cls(e) if (e := _load_enum(cls._Enum, custom)) else None


H = OneQbGate(OneQbGate._Enum.H)


@dataclass(frozen=True)
class TwoQbGate(AsExtOp):
    class _Enum(Enum):
        CX = "CX"

    _enum: _Enum

    def op_def(self) -> ext.OpDef:
        return QUANTUM_EXT.operations[self._enum.value]

    @classmethod
    def from_ext(cls, custom: ExtOp) -> Self | None:
        return cls(e) if (e := _load_enum(cls._Enum, custom)) else None

    def __call__(self, q0: ComWire, q1: ComWire) -> Command:
        return DataflowOp.__call__(self, q0, q1)


CX = TwoQbGate(TwoQbGate._Enum.CX)


@QUANTUM_EXT.register_op(
    "Measure",
    signature=tys.FunctionType([tys.Qubit], [tys.Qubit, tys.Bool]),
)
@dataclass(frozen=True)
class MeasureDef(RegisteredOp):
    def __call__(self, q: ComWire) -> Command:
        return super().__call__(q)


Measure = MeasureDef()


@QUANTUM_EXT.register_op(
    "MeasureFree",
    signature=tys.FunctionType([tys.Qubit], [tys.Bool]),
)
@dataclass(frozen=True)
class MeasureFreeDef(RegisteredOp):
    def __call__(self, q: ComWire) -> Command:
        return super().__call__(q)


MeasureFree = MeasureFreeDef()


@QUANTUM_EXT.register_op(
    "QAlloc",
    signature=tys.FunctionType([], [tys.Qubit]),
)
@dataclass(frozen=True)
class QAllocDef(RegisteredOp):
    def __call__(self) -> Command:
        return super().__call__()


QAlloc = QAllocDef()


@QUANTUM_EXT.register_op(
    "Rz",
    signature=tys.FunctionType([tys.Qubit, FLOAT_T], [tys.Qubit]),
)
@dataclass(frozen=True)
class RzDef(RegisteredOp):
    def __call__(self, q: ComWire, fl_wire: ComWire) -> Command:
        return super().__call__(q, fl_wire)


Rz = RzDef()


def validate(
    h: Hugr | Package,
    *,
    roundtrip: bool = True,
    snap: SnapshotAssertion | None = None,
):
    """Validate a HUGR or package.

    args:
        h: The HUGR or package to validate.
        roundtrip: Whether to roundtrip the HUGR through the CLI.
        snapshot: A hugr render snapshot. If not None, it will be compared against the
        rendered HUGR. Pass `--snapshot-update` to pytest to update the snapshot file.
    """
    if snap is not None:
        dot = h.render_dot() if isinstance(h, Hugr) else h.modules[0].render_dot()
        assert snap == dot.source
        if os.environ.get("HUGR_RENDER_DOT"):
            dot.pipe("svg")

    # Encoding formats to test. Note that these include other formats than
    # those supported by `hugr convert`.
    FORMATS = {
        "json": EnvelopeConfig.TEXT,
        "json-compressed": EnvelopeConfig(format=EnvelopeFormat.JSON, zstd=0),
        "model-exts": EnvelopeConfig.BINARY,
        "model-exts-no-compression": EnvelopeConfig(
            format=EnvelopeFormat.MODEL_WITH_EXTS, zstd=None
        ),
        "s-expression": EnvelopeConfig(format=EnvelopeFormat._S_EXPRESSION, zstd=None),
        "s-expression-exts": EnvelopeConfig(
            format=EnvelopeFormat._S_EXPRESSION_WITH_EXTS, zstd=None
        ),
    }
    # Envelope formats used when exporting test hugrs.
    WRITE_FORMATS = [
        "json",
        "json-compressed",
        "model-exts",
        "model-exts-no-compression",
        # Requires extension-encoding for the cli validation to load the envelope
        # "s-expression",
        # TODO: Raises multiple errors when loading values
        # "s-expression-exts",
    ]
    # Envelope formats used as target before loading back the test hugrs.
    # These should correspond to the formats supported by `hugr convert`.
    LOAD_FORMATS = [
        "json",
        "model-exts",
        # TODO: Raises multiple errors when loading values
        # "s-expression-exts"
    ]

    # validate text and binary formats
    for write_fmt in WRITE_FORMATS:
        serial = h.to_bytes(FORMATS[write_fmt])
        cli.validate(serial)

        if roundtrip:
            # Roundtrip tests:
            # Try converting to all possible LOAD_FORMATS, load them back in,
            # and check that the loaded HUGR corresponds to the original using
            # a node hash comparison.
            #
            # Run `pytest` with `-vv` to see the hash diff.
            for load_fmt in LOAD_FORMATS:
                if load_fmt != write_fmt:
                    converted_serial = cli.convert(serial, format=load_fmt)
                else:
                    converted_serial = serial
                loaded = Package.from_bytes(converted_serial)

                modules = [h] if isinstance(h, Hugr) else h.modules

                assert len(loaded.modules) == len(modules)
                for m1, m2 in zip(loaded.modules, modules, strict=True):
                    h1_hash = _NodeHash.hash_hugr(m1, "original")
                    h2_hash = _NodeHash.hash_hugr(m2, "loaded")
                    assert (
                        h1_hash == h2_hash
                    ), f"HUGRs are not the same for {write_fmt} -> {load_fmt}"


def canonicalize_json(payload):
    """Put json into a canonical form for hashing purposes.

    Specifically, replace a general sum of empty rows with an explicit unit sum."""
    if isinstance(payload, list):
        return list(map(canonicalize_json, payload))
    elif isinstance(payload, dict):
        if (
            set(payload.keys()) == {"t", "s", "rows"}
            and payload["t"] == "Sum"
            and payload["s"] == "General"
            and not any(payload["rows"])
        ):
            return {"t": "Sum", "s": "Unit", "size": len(payload["rows"])}
        else:
            return {k: canonicalize_json(v) for k, v in payload.items()}
    else:
        return payload


@dataclass(frozen=True, order=True)
class _NodeHash:
    op: _OpHash
    entrypoint: bool
    input_neighbours: int
    output_neighbours: int
    input_ports: int
    output_ports: int
    input_order_edges: int
    output_order_edges: int
    is_region: bool
    node_depth: int
    children_hashes: list[_NodeHash]
    metadata: dict[str, str]

    @classmethod
    def hash_hugr(cls, h: Hugr, name: str) -> _NodeHash:
        """Returns an order-independent hash of a HUGR."""
        return cls._hash_node(h, h.module_root, 0, name)

    @classmethod
    def _hash_node(cls, h: Hugr, n: Node, depth: int, name: str) -> _NodeHash:
        children = h.children(n)
        child_hashes = sorted(cls._hash_node(h, c, depth + 1, name) for c in children)
        metadata = {k: str(v) for k, v in h[n].metadata.items()}

        # Pick a normalized representation of the op name.
        op_type = h[n].op
        if isinstance(op_type, AsExtOp):
            op_type = op_type.ext_op.to_custom_op()
            op = _OpHash(f"{op_type.extension}.{op_type.op_name}")
        elif isinstance(op_type, Custom):
            op = _OpHash(f"{op_type.extension}.{op_type.op_name}")
        elif isinstance(op_type, Const):
            # We need every custom value to have the same repr if they compare
            # equal. For example, an `IntVal(42)` should be the same as the
            # equivalent `Extension` value. This needs a lot of extra
            # unwrapping, since each class implements different `__repr__`
            # methods.
            #
            # Our solution here is to encode the value into JSON and compare those.
            # This may miss some errors, but it's the best we can do for now. Note that
            # roundtripping via `sops.Value` is not enough, since nested
            # specialized values don't get serialized straight away. (e.g.
            # StaticArrayVal's dictionary payload containing a SumValue
            # internally, see `test_val_static_array`).
            value_dict = op_type.val._to_serial_root().model_dump(mode="json")
            op = _OpHash("Const", canonicalize_json(value_dict))
        else:
            op = _OpHash(op_type.name())

        return _NodeHash(
            entrypoint=n == h.entrypoint,
            op=op,
            input_neighbours=h.num_incoming(n),
            output_neighbours=h.num_outgoing(n),
            input_ports=h.num_in_ports(n),
            output_ports=h.num_out_ports(n),
            input_order_edges=len(list(h.incoming_order_links(n))),
            output_order_edges=len(list(h.outgoing_order_links(n))),
            is_region=len(children) > 0,
            node_depth=depth,
            children_hashes=child_hashes,
            metadata=metadata,
        )


@dataclass(frozen=True)
class _OpHash:
    name: str
    payload: None | typing.Any = None

    def __lt__(self, other: _OpHash) -> bool:
        """Compare two op hashes by name and payload."""
        return (self.name, repr(self.payload)) < (other.name, repr(other.payload))



"""HUGR integer types and operations."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, ClassVar

from typing_extensions import Self

import hugr.model as model
from hugr import ext, tys, val
from hugr.ops import AsExtOp, DataflowOp, ExtOp, RegisteredOp
from hugr.std._util import _load_extension

if TYPE_CHECKING:
    from hugr.ops import Command, ComWire

CONVERSIONS_EXTENSION = _load_extension("arithmetic.conversions")

INT_TYPES_EXTENSION = _load_extension("arithmetic.int.types")
_INT_PARAM = tys.BoundedNatParam(7)

INT_T_DEF = INT_TYPES_EXTENSION.types["int"]


def int_t(width: int) -> tys.ExtType:
    """Create an integer type with a fixed log bit width.


    Args:
        width: The log bit width of the integer.

    Returns:
        The integer type.

    Examples:
        >>> int_t(5).type_def.name # 32 bit integer
        'int'
    """
    return INT_T_DEF.instantiate(
        [tys.BoundedNatArg(n=width)],
    )


def _int_tv(index: int) -> tys.ExtType:
    return INT_T_DEF.instantiate(
        [tys.VariableArg(idx=index, param=_INT_PARAM)],
    )


#: HUGR 32-bit integer type.
INT_T = int_t(5)


def _to_unsigned(val: int, bits: int) -> int:
    """Convert a signed integer to its unsigned representation
    in twos-complement form.

    Positive integers are unchanged, while negative integers
    are converted by adding 2^bits to the value.

    Raises ValueError if the value is out of range for the given bit width
    (valid range is [-2^(bits-1), 2^(bits-1)-1]).
    """
    half_max = 1 << (bits - 1)
    min_val = -half_max
    max_val = half_max - 1
    if val < min_val or val > max_val:
        msg = f"Value {val} out of range for {bits}-bit signed integer."
        raise ValueError(msg)  #

    if val < 0:
        return (1 << bits) + val
    return val


@dataclass
class IntVal(val.ExtensionValue):
    """Custom value for a signed integer."""

    v: int
    width: int = field(default=5)

    def to_value(self) -> val.Extension:
        name = "ConstInt"
        unsigned = _to_unsigned(self.v, 1 << self.width)
        payload = {"log_width": self.width, "value": unsigned}
        return val.Extension(
            name,
            typ=int_t(self.width),
            val=payload,
        )

    def __str__(self) -> str:
        return f"{self.v}"

    def _resolve_used_extensions_inplace(
        self, registry: ext.ExtensionRegistry | None = None
    ) -> ext.ExtensionResolutionResult:
        result = ext.ExtensionResolutionResult()
        result.used_extensions.add_extension(INT_TYPES_EXTENSION)
        return result

    def to_model(self) -> model.Term:
        unsigned = _to_unsigned(self.v, 1 << self.width)
        return model.Apply(
            "arithmetic.int.const", [model.Literal(self.width), model.Literal(unsigned)]
        )


INT_OPS_EXTENSION = _load_extension("arithmetic.int")


@dataclass(frozen=True)
class _DivModDef(RegisteredOp):
    """DivMod operation, has two inputs and two outputs."""

    width: int = 5
    const_op_def: ClassVar[ext.OpDef] = INT_OPS_EXTENSION.operations["idivmod_u"]

    def type_args(self) -> list[tys.TypeArg]:
        return [tys.BoundedNatArg(n=self.width)]

    def cached_signature(self) -> tys.FunctionType | None:
        row: list[tys.Type] = [int_t(self.width)] * 2
        return tys.FunctionType.endo(row)

    @classmethod
    def from_ext(cls, custom: ExtOp) -> Self | None:
        if custom.op_def() != cls.op_def():
            return None
        match custom.args:
            case [tys.BoundedNatArg(n=a1)]:
                return cls(width=a1)
            case _:
                msg = f"Invalid args: {custom.args}"
                raise AsExtOp.InvalidExtOp(msg)

    def __call__(self, a: ComWire, b: ComWire) -> Command:
        return DataflowOp.__call__(self, a, b)


#: DivMod operation.
DivMod = _DivModDef()



"""Borrow array types and operations."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, cast

import hugr.model as model
from hugr import tys, val
from hugr.std._util import _load_extension
from hugr.utils import comma_sep_str

if TYPE_CHECKING:
    from hugr.ext import ExtensionRegistry, ExtensionResolutionResult

EXTENSION = _load_extension("collections.borrow_arr")


@dataclass(eq=False)
class BorrowArray(tys.ExtType):
    """Fixed `size` borrow array of `ty` elements."""

    def __init__(self, ty: tys.Type, size: int | tys.TypeArg) -> None:
        if isinstance(size, int):
            size = tys.BoundedNatArg(size)

        err_msg = (
            f"Borrow array size must be a bounded natural or a nat variable, not {size}"
        )
        match size:
            case tys.BoundedNatArg(_n):
                pass
            case tys.VariableArg(_idx, param):
                if not isinstance(param, tys.BoundedNatParam):
                    raise ValueError(err_msg)  # noqa: TRY004
            case _:
                raise ValueError(err_msg)

        ty_arg = tys.TypeTypeArg(ty)

        self.type_def = EXTENSION.types["borrow_array"]
        self.args = [size, ty_arg]

    @property
    def ty(self) -> tys.Type:
        assert isinstance(
            self.args[1], tys.TypeTypeArg
        ), "Borrow array elements must have a valid type"
        return self.args[1].ty

    @property
    def size(self) -> int | None:
        """If the borrow array has a concrete size, return it.

        Otherwise, return None.
        """
        if isinstance(self.args[0], tys.BoundedNatArg):
            return self.args[0].n
        return None

    def type_bound(self) -> tys.TypeBound:
        return tys.TypeBound.Linear

    def _resolve_used_extensions(
        self, registry: ExtensionRegistry | None = None
    ) -> tuple[BorrowArray, ExtensionResolutionResult]:
        ext_type, result = super()._resolve_used_extensions(registry)

        assert isinstance(
            ext_type, tys.ExtType
        ), "HUGR internal error, expected resolved type to be extension type."
        assert (
            ext_type.type_def == EXTENSION.types["borrow_array"]
        ), "HUGR internal error, expected resolved type to be borrow array."

        borrow_array = BorrowArray(tys.Unit, 0)
        borrow_array.args = ext_type.args
        return borrow_array, result


# Note that only borrow array values with no elements borrowed should be emitted.
@dataclass
class BorrowArrayVal(val.ExtensionValue):
    """Constant value for a statically sized borrow array of elements."""

    v: list[val.Value]
    ty: BorrowArray

    def __init__(self, v: list[val.Value], elem_ty: tys.Type) -> None:
        self.v = v
        self.ty = BorrowArray(elem_ty, len(v))

    def to_value(self) -> val.Extension:
        name = "BorrowArrayValue"
        # The value list must be serialized at this point, otherwise the
        # `Extension` value would not be serializable.
        vs = [v._to_serial_root() for v in self.v]
        element_ty = self.ty.ty._to_serial_root()
        serial_val = {"values": vs, "typ": element_ty}
        return val.Extension(name, typ=self.ty, val=serial_val)

    def __str__(self) -> str:
        return f"borrow_array({comma_sep_str(self.v)})"

    def to_model(self) -> model.Term:
        return model.Apply(
            "collections.borrow_array.const",
            [
                model.Literal(len(self.v)),
                cast(model.Term, self.ty.ty.to_model()),
                model.List([value.to_model() for value in self.v]),
            ],
        )

    def _resolve_used_extensions_inplace(
        self, registry: ExtensionRegistry | None = None
    ) -> ExtensionResolutionResult:
        resolved_ty, result = self.ty._resolve_used_extensions(registry)
        assert isinstance(
            resolved_ty, BorrowArray
        ), "HUGR internal error, expected resolved type to be borrow array."
        self.ty = resolved_ty
        for value in self.v:
            result.extend(value._resolve_used_extensions_inplace(registry))
        return result



"""Core data structures for HUGR."""

from __future__ import annotations

import copy
import json
from collections.abc import Iterable, Iterator, Mapping
from dataclasses import dataclass, field
from queue import Queue
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    TypeVar,
    cast,
    overload,
)

from typing_extensions import deprecated

import hugr.model as model
import hugr.ops as ops
from hugr._serialization.ops import OpType as SerialOp
from hugr._serialization.serial_hugr import SerialHugr
from hugr.envelope import (
    EnvelopeConfig,
    _make_envelope,
    _make_envelope_str,
    read_envelope_hugr,
    read_envelope_hugr_str,
)
from hugr.exceptions import ParentBeforeChild
from hugr.metadata import NodeMetadata
from hugr.ops import (
    CFG,
    Call,
    Conditional,
    Const,
    DataflowBlock,
    DataflowOp,
    ExitBlock,
    FuncDefn,
    IncompleteOp,
    Module,
    Op,
    is_dataflow_op,
)
from hugr.tys import Kind, OrderKind, Type, ValueKind
from hugr.utils import BiMap
from hugr.val import Value

from .node_port import (
    Direction,
    InPort,
    Node,
    NodeIdx,
    OutPort,
    PortOffset,
    ToNode,
    _SubPort,
)

if TYPE_CHECKING:
    import graphviz as gv  # type: ignore[import-untyped]

    from hugr import ext
    from hugr.ext import ExtensionRegistry
    from hugr.package import Package
    from hugr.val import Value

    from .render import RenderConfig


@dataclass()
class NodeData:
    """Node weights in HUGR graph. Defined by an operation and parent node."""

    #: The operation of the node.
    op: Op
    #: The parent node, or None for the root node.
    parent: Node | None
    _num_inps: int = field(default=0, repr=False)
    _num_outs: int = field(default=0, repr=False)
    children: list[Node] = field(default_factory=list, repr=False)
    metadata: NodeMetadata = field(default_factory=NodeMetadata)

    def _to_serial(self, node: Node) -> SerialOp:
        o = self.op._to_serial(self.parent if self.parent else node)

        return SerialOp(root=o)  # type: ignore[arg-type]


_SO = _SubPort[OutPort]
_SI = _SubPort[InPort]

P = TypeVar("P", InPort, OutPort)
K = TypeVar("K", InPort, OutPort)
OpVar = TypeVar("OpVar", bound=Op)
OpVarCov = TypeVar("OpVarCov", bound=Op, covariant=True)


@dataclass()
class Hugr(Mapping[Node, NodeData], Generic[OpVarCov]):
    """The core HUGR datastructure.

    Args:
        entrypoint_op: The operation for the entrypoint node. Defaults to a Module
        (which will then also be the root).

    Examples:
        >>> h = Hugr()
        >>> h.entrypoint_op()
        Module()
        >>> h[h.entrypoint].op
        Module()
        >>> dfg_h = Hugr(ops.DFG([tys.Bool]))
        >>> dfg_h[dfg_h.entrypoint].op
        DFG(inputs=[Bool])
    """

    # The module at the root of the HUGR.
    module_root: Node
    # Entrypoint node for the HUGR.
    #
    # Most traversals and rewrite operations start from this node.
    #
    # This node may be of any optype that's the parent to a region, and is a
    # descendant of the module definition at the HUGR root (or the root itself).
    entrypoint: Node
    # List of nodes, with None for deleted nodes.
    _nodes: list[NodeData | None]
    # Bidirectional map of links between ports.
    _links: BiMap[_SO, _SI]
    # List of free node indices, populated when nodes are deleted.
    _free_nodes: list[Node]

    def __init__(self, entrypoint_op: OpVarCov | None = None) -> None:
        self._free_nodes = []
        self._links = BiMap()
        self._nodes = []
        self.entrypoint = Node(0)
        self.module_root = Node(0)

        # The root of a HUGR is always a module.
        self.module_root = self._add_node(Module(), None, 0)
        self.entrypoint = self.module_root

        unsupported_op_msg = (
            f"Creating new HUGRs with entrypoint {entrypoint_op} is not supported"
        )

        # Depending on the entrypoint op, we may need to
        # wrap nest it inside the root module.
        match entrypoint_op:
            case None | Module():
                pass
            case ops.FuncDefn():
                self.entrypoint = self.add_node(
                    entrypoint_op, self.module_root, num_outs=1
                )
            case _:
                from hugr.build import Function

                # Some operations are unsupported, as they require additional context to
                # be valid (e.g. cfg blocks, case statements, etc.).
                if not ops.is_dataflow_op(entrypoint_op):
                    raise ValueError(unsupported_op_msg)
                # Explicit type required to keep mypy happy
                df_op: ops.DataflowOp = entrypoint_op

                # Only region containers are allowed to be entrypoints
                match df_op:
                    case CFG() | Conditional():
                        pass
                    case _ if ops.is_df_parent_op(df_op):
                        pass
                    case _:
                        raise ValueError(unsupported_op_msg)
                df_op = cast(ops.DataflowOp, df_op)  # checked above, to appease mypy
                inputs, outputs = None, None
                try:
                    sig = df_op.outer_signature()
                    inputs = sig.input
                    outputs = sig.output
                except IncompleteOp:
                    match df_op:
                        case CFG():
                            inputs = df_op.inputs
                        case _:
                            inputs = df_op._inputs()

                parent_op = FuncDefn("main", inputs, [])
                func = Function.new_nested(parent_op, self, self.module_root)

                if outputs is not None:
                    self.entrypoint = func.add_op(df_op, *func.inputs())
                    func.set_outputs(*self.entrypoint.outputs())
                else:
                    # Connecting the entrypoint to the function's output is delayed
                    # until `set_outputs` is called.
                    # See `hugr._connect_df_entrypoint_outputs`.
                    self.entrypoint = self.add_node(df_op, func)
                    func._wire_up(self.entrypoint, func.inputs())
                    df_op._entrypoint_requires_wiring = True

    def __getitem__(self, key: ToNode) -> NodeData:
        key = key.to_node()
        try:
            n = self._nodes[key.idx]
        except IndexError:
            n = None
        if n is None:
            raise KeyError(key)
        return n

    def __iter__(self) -> Iterator[Node]:
        return (
            Node(idx, data.metadata)
            for idx, data in enumerate(self._nodes)
            if data is not None
        )

    def __len__(self) -> int:
        return self.num_nodes()

    def _get_typed_op(self, node: ToNode, cl: type[OpVar]) -> OpVar:
        op = self[node].op
        assert isinstance(op, cl)
        return op

    def nodes(self) -> Iterable[tuple[Node, NodeData]]:
        """Iterator over all the nodes of the hugr and their data.

        To get the descendants of the entrypoint, use `descendants()`.
        """
        return self.items()

    def sorted_region_nodes(self, parent: Node) -> Iterator[Node]:
        """Iterator over a topological ordering of all the hugr nodes.

        Note that the sort is performed within a hugr region and non-local
        edges are ignored.

        Args:
            parent: The parent node of the region to sort.

        Raises:
            ValueError: If the region contains a cycle.

        Examples:
            >>> from hugr.build.tracked_dfg import TrackedDfg
            >>> from hugr.std.logic import Not
            >>> dfg = TrackedDfg(tys.Bool)
            >>> [b] = dfg.track_inputs()
            >>> for _ in range(6):
            ...     _= dfg.add(Not(b));
            >>> dfg.set_tracked_outputs()
            >>> nodes = list(dfg.hugr)
            >>> list(dfg.hugr.sorted_region_nodes(nodes[4]))
            [Node(5), Node(7), Node(8), Node(9), Node(10), Node(11), Node(12), Node(6)]
        """
        # A dict to keep track of how many times we see a node.
        # Store the Nodes with the input degrees as values.
        # Implementation uses Kahn's algorithm
        # https://en.wikipedia.org/wiki/Topological_sorting#Kahn's_algorithm
        visit_dict: dict[Node, int] = {}
        queue: Queue[Node] = Queue()
        for node in self.children(parent):
            incoming = 0
            for n in self.input_neighbours(node):
                same_region = self[n].parent == parent
                # Only update the degree of the node if edge is within the same region.
                # We do not count non-local edges.
                if same_region:
                    incoming += 1
            if incoming:
                visit_dict[node] = incoming
            # If a Node has no dependencies, add it to the queue.
            else:
                queue.put(node)

        while not queue.empty():
            new_node = queue.get()
            yield new_node

            for neigh in self.output_neighbours(new_node):
                visit_dict[neigh] -= 1
                if visit_dict[neigh] == 0:
                    del visit_dict[neigh]
                    queue.put(neigh)

        # If our dict is non-empty here then our graph contains a cycle
        if visit_dict:
            err = "Graph contains a cycle. No topological ordering exists."
            raise ValueError(err)

    def links(self) -> Iterator[tuple[OutPort, InPort]]:
        """Iterator over all the links in the HUGR.

        Returns:
            Iterator of pairs of outgoing port and the incoming ports.
        """
        return ((src.port, tgt.port) for src, tgt in self._links.items())

    def children(self, node: ToNode | None = None) -> list[Node]:
        """The child nodes of a given `node`.

        Args:
            node: Parent node. Defaults to the HUGR entrypoint.

        Returns:
            List of child nodes.

        Examples:
            >>> h = Hugr()
            >>> n = h.add_node(ops.Const(val.TRUE))
            >>> h.children(h.entrypoint)
            [Node(1)]
        """
        node = node or self.entrypoint
        return self[node].children

    def descendants(self, node: ToNode | None = None) -> Iterable[Node]:
        """Iterator over all the descendants of the hugr entrypoint.

        Traverses the HUGR graph in a breadth-first manner, starting from
        the entrypoint.

        To get all the nodes in the HUGR, use `nodes()`.

        Args:
            node: Parent node. Defaults to the HUGR entrypoint.

        Returns:
            List of child nodes.
        """
        queue: Queue[Node] = Queue()
        queue.put(node.to_node() if node is not None else self.entrypoint)
        while not queue.empty():
            node = queue.get()
            yield node
            for child in self.children(node):
                queue.put(child)

    def _add_node(
        self,
        op: Op,
        parent: ToNode | None = None,
        num_outs: int | None = None,
        metadata: dict[str, Any] | NodeMetadata | None = None,
    ) -> Node:
        parent = parent.to_node() if parent else None

        if metadata is None or isinstance(metadata, dict):
            metadata = NodeMetadata(metadata)

        node_data = NodeData(op, parent, metadata=metadata)

        if self._free_nodes:
            node = self._free_nodes.pop()
            self._nodes[node.idx] = node_data
        else:
            node = Node(len(self._nodes), NodeMetadata())
            self._nodes.append(node_data)
        node._num_out_ports = num_outs
        node._metadata = node_data.metadata
        if parent:
            self[parent].children.append(node)

        self._update_node_outs(node, num_outs)
        return node

    def _update_node_outs(self, node: Node, num_outs: int | None) -> Node:
        """Update the number of outgoing ports for a node.

        Returns:
            The updated node.
        """
        return self._update_port_count(node, num_outs=num_outs)

    def _update_port_count(
        self, node: Node, *, num_inps: int | None = None, num_outs: int | None
    ) -> Node:
        """Update the number of incoming and outgoing ports for a node.

        If `num_inps` or `num_outs` is None, the corresponding count is not updated.

        Returns:
            The updated node.
        """
        if num_inps is None and num_outs is None:
            return node

        if num_inps is not None:
            self[node]._num_inps = num_inps
        if num_outs is not None:
            self[node]._num_outs = num_outs
            node._num_out_ports = num_outs

        if node.idx == self.entrypoint.idx:
            self.entrypoint = node
        if node.idx == self.module_root.idx:
            self.module_root = node

        return node

    def add_node(
        self,
        op: Op,
        parent: ToNode | None = None,
        num_outs: int | None = None,
        metadata: dict[str, Any] | NodeMetadata | None = None,
    ) -> Node:
        """Add a node to the HUGR.

        Args:
            op: Operation of the node.
            parent: Parent node of added node. Defaults to HUGR entrypoint if None.
            num_outs: Number of output ports expected for this node. Defaults to None.
            metadata: A dictionary of metadata to associate with the node.
                Defaults to None.

        Returns:
            Handle to the added node.
        """
        parent = parent or self.entrypoint
        return self._add_node(op, parent, num_outs, metadata)

    def add_const(
        self,
        value: Value,
        parent: ToNode | None = None,
        metadata: dict[str, Any] | NodeMetadata | None = None,
    ) -> Node:
        """Add a constant node to the HUGR.

        Args:
            value: Value of the constant.
            parent: Parent node of added node. Defaults to HUGR entrypoint if None.
            metadata: A dictionary of metadata to associate with the node.
                Defaults to None.

        Returns:
            Handle to the added node.

        Examples:
            >>> h = Hugr()
            >>> n = h.add_const(val.TRUE)
            >>> h[n].op
            Const(TRUE)
        """
        return self.add_node(Const(value), parent, metadata=metadata)

    def delete_node(self, node: ToNode) -> NodeData | None:
        """Delete a node from the HUGR.

        Args:
            node: Node to delete.

        Returns:
            The deleted node data, or None if the node was not found.

        Examples:
            >>> h = Hugr()
            >>> n = h.add_const(val.TRUE)
            >>> deleted = h.delete_node(n)
            >>> deleted.op
            Const(TRUE)
            >>> len(h)
            1
        """
        node = node.to_node()
        parent = self[node].parent
        if parent:
            self[parent].children.remove(node)
        for inp, _ in self.incoming_links(node):
            self._links.delete_right(_SubPort(inp))
        for out, _ in self.outgoing_links(node):
            self._links.delete_left(_SubPort(out))

        weight, self._nodes[node.idx] = self._nodes[node.idx], None

        # Free up the metadata dictionary
        node._metadata = NodeMetadata()

        self._free_nodes.append(node)
        return weight

    def _unused_sub_offset(self, port: P) -> _SubPort[P]:
        d: dict[_SO, _SI] | dict[_SI, _SO]
        match port:
            case OutPort(_):
                d = self._links.fwd
            case InPort(_):
                d = self._links.bck
        sub_port = _SubPort(port)
        while sub_port in d:
            sub_port = sub_port.next_sub_offset()
        return sub_port

    def has_link(self, src: OutPort, dst: InPort) -> bool:
        """Check if there is a link between two ports.

        Args:
            src: Source port.
            dst: Destination port.

        Returns:
            True if there is a link, False otherwise.

        Examples:
            >>> df = dfg.Dfg(tys.Bool)
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(0))
            >>> df.hugr.has_link(df.input_node.out(0), df.output_node.inp(0))
            True
        """
        return dst in self.linked_ports(src)

    def add_link(self, src: OutPort, dst: InPort) -> None:
        """Add a link (edge) between two nodes to the HUGR,
          from an outgoing port to an incoming port.

        Args:
            src: Source port.
            dst: Destination port.

        Examples:
            >>> df = dfg.Dfg(tys.Bool)
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(0))
            >>> list(df.hugr.linked_ports(df.input_node[0]))
            [InPort(Node(6), 0)]
        """
        src_sub = self._unused_sub_offset(src)
        dst_sub = self._unused_sub_offset(dst)
        # if self._links.get_left(dst_sub) is not None:
        #     dst = replace(dst, _sub_offset=dst._sub_offset + 1)
        self._links.insert_left(src_sub, dst_sub)

        self[src.node]._num_outs = max(self[src.node]._num_outs, src.offset + 1)
        self[dst.node]._num_inps = max(self[dst.node]._num_inps, dst.offset + 1)

    def add_order_link(self, src: ToNode, dst: ToNode) -> None:
        """Add a state order link between two nodes.

        Args:
            src: Source node.
            dst: Destination node.

        Examples:
            >>> df = dfg.Dfg()
            >>> df.hugr.add_order_link(df.input_node, df.output_node)
            >>> list(df.hugr.outgoing_order_links(df.input_node))
            [Node(6)]
        """
        source = src.out(-1)
        target = dst.inp(-1)
        assert (
            self.port_kind(source) == OrderKind()
        ), f"Operation {self[src].op.name()} does not support order edges"
        assert (
            self.port_kind(target) == OrderKind()
        ), f"Operation {self[dst].op.name()} does not support order edges"
        if not self.has_link(source, target):
            self.add_link(source, target)

    def delete_link(self, src: OutPort, dst: InPort) -> None:
        """Delete a link (edge) between two nodes from the HUGR.

        Args:
            src: Source port.
            dst: Destination port.
        """
        try:
            sub_offset = next(
                i for i, inp in enumerate(self.linked_ports(src)) if inp == dst
            )
            self._links.delete_left(_SubPort(src, sub_offset))
        except StopIteration:
            return
        # TODO make sure sub-offset is handled correctly

    def entrypoint_op(self) -> OpVarCov:
        """The operation of the root node.

        Examples:
            >>> h = Hugr()
            >>> h.entrypoint_op()
            Module()
        """
        return cast("OpVarCov", self[self.entrypoint].op)

    def num_nodes(self) -> int:
        """The number of nodes in the HUGR.

        Examples:
            >>> h = Hugr()
            >>> n = h.add_const(val.TRUE)
            >>> h.num_nodes()
            2
        """
        return len(self._nodes) - len(self._free_nodes)

    def num_ports(self, node: ToNode, direction: Direction) -> int:
        """The number of ports of a node in a given direction.
        Not necessarily the number of connected ports - if port `i` is
        connected, then all ports `0..i` are assumed to exist.

        This value includes order ports.

        Args:
            node: Node to query.
            direction: Direction of ports to count.

        Examples:
            >>> from hugr.std.logic import Not
            >>> h = Hugr()
            >>> n1 = h.add_node(Not)
            >>> n2 = h.add_node(Not)
            >>> # Passing offset `2` here allocates new ports automatically
            >>> h.add_link(n1.out(0), n2.inp(2))
            >>> h.add_order_link(n1, n2)
            >>> h.num_ports(n1, Direction.OUTGOING)
            1
            >>> h.num_ports(n2, Direction.INCOMING)
            3
        """
        return (
            self.num_in_ports(node)
            if direction == Direction.INCOMING
            else self.num_out_ports(node)
        )

    def num_in_ports(self, node: ToNode) -> int:
        """The number of incoming ports of a node. See :meth:`num_ports`.

        This value does not include order ports.
        """
        return self[node]._num_inps

    def num_out_ports(self, node: ToNode) -> int:
        """The number of outgoing ports of a node. See :meth:`num_ports`.

        This value does not include order ports.
        """
        return self[node]._num_outs

    def _linked_ports(
        self, port: P, links: dict[_SubPort[P], _SubPort[K]]
    ) -> Iterable[K]:
        sub_port = _SubPort(port)
        while sub_port in links:
            # sub offset not used in API
            yield links[sub_port].port
            sub_port = sub_port.next_sub_offset()

    @overload
    def linked_ports(self, port: OutPort) -> Iterable[InPort]: ...
    @overload
    def linked_ports(self, port: InPort) -> Iterable[OutPort]: ...
    def linked_ports(self, port: OutPort | InPort):
        """Return an iterable of In(Out)Ports linked to given Out(In)Port.

        Args:
            port: Given port.

        Returns:
            Iterator over linked ports.

        Examples:
            >>> df = dfg.Dfg(tys.Bool)
            >>> df.set_outputs(df.input_node[0])
            >>> list(df.hugr.linked_ports(df.input_node[0]))
            [InPort(Node(6), 0)]

        """
        match port:
            case OutPort(_):
                return self._linked_ports(port, self._links.fwd)
            case InPort(_):
                return self._linked_ports(port, self._links.bck)

    # TODO: single linked port

    def outgoing_order_links(self, node: ToNode) -> Iterable[Node]:
        """Iterator over nodes connected by an outgoing state order link from a
        given node.

        Args:
            node: Source node of state order link.

        Examples:
            >>> df = dfg.Dfg()
            >>> df.add_state_order(df.input_node, df.output_node)
            >>> list(df.hugr.outgoing_order_links(df.input_node))
            [Node(6)]
        """
        return (p.node for p in self.linked_ports(node.out(-1)))

    def incoming_order_links(self, node: ToNode) -> Iterable[Node]:
        """Iterator over nodes connected by an incoming state order link to a
        given node.

        Args:
            node: Destination node of state order link.

        Examples:
            >>> df = dfg.Dfg()
            >>> df.add_state_order(df.input_node, df.output_node)
            >>> list(df.hugr.incoming_order_links(df.output_node))
            [Node(5)]
        """
        return (p.node for p in self.linked_ports(node.inp(-1)))

    def _node_links(
        self, node: ToNode, links: dict[_SubPort[P], _SubPort[K]]
    ) -> Iterable[tuple[P, list[K]]]:
        try:
            direction = next(iter(links.keys())).port.direction
        except StopIteration:
            return
        # iterate over known offsets
        for offset in range(self.num_ports(node, direction)):
            port = cast("P", node.port(offset, direction))
            yield port, list(self._linked_ports(port, links))

        order_port = cast("P", node.port(-1, direction))
        linked_order = list(self._linked_ports(order_port, links))
        if linked_order:
            yield order_port, linked_order

    def outgoing_links(self, node: ToNode) -> Iterable[tuple[OutPort, list[InPort]]]:
        """Iterator over outgoing links from a given node.

        This number includes order ports.

        Args:
            node: Node to query.

        Returns:
            Iterator of pairs of outgoing port and the incoming ports connected
            to that port.

        Examples:
            >>> df = dfg.Dfg()
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(0))
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(1))
            >>> df.hugr.add_order_link(df.input_node, df.output_node)
            >>> list(df.hugr.outgoing_links(df.input_node))
            [(OutPort(Node(5), 0), [InPort(Node(6), 0), InPort(Node(6), 1)]), (OutPort(Node(5), -1), [InPort(Node(6), -1)])]
        """  # noqa: E501
        return self._node_links(node, self._links.fwd)

    def incoming_links(self, node: ToNode) -> Iterable[tuple[InPort, list[OutPort]]]:
        """Iterator over incoming links to a given node.

        This number includes order ports.

        Args:
            node: Node to query.

        Returns:
            Iterator of pairs of incoming port and the outgoing ports connected
            to that port.

        Examples:
            >>> df = dfg.Dfg()
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(0))
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(1))
            >>> df.hugr.add_order_link(df.input_node, df.output_node)
            >>> list(df.hugr.incoming_links(df.output_node))
            [(InPort(Node(6), 0), [OutPort(Node(5), 0)]), (InPort(Node(6), 1), [OutPort(Node(5), 0)]), (InPort(Node(6), -1), [OutPort(Node(5), -1)])]
        """  # noqa: E501
        return self._node_links(node, self._links.bck)

    def neighbours(
        self, node: ToNode, direction: Direction | None = None
    ) -> Iterable[Node]:
        """Iterator over the neighbours of a node.

        Args:
            node: Node to query.
            direction: If given, only return neighbours in that direction.

        Returns:
            Iterator of nodes connected to `node`, ordered by direction and port
            offset. Nodes connected via multiple links will be returned multiple times.

        Examples:
            >>> df = dfg.Dfg()
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(0))
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(1))
            >>> list(df.hugr.neighbours(df.input_node))
            [Node(6), Node(6)]
            >>> list(df.hugr.neighbours(df.output_node, Direction.OUTGOING))
            []
        """
        if direction is None or direction == Direction.INCOMING:
            for _, linked_outputs in self.incoming_links(node):
                for out_port in linked_outputs:
                    yield out_port.node
        if direction is None or direction == Direction.OUTGOING:
            for _, linked_inputs in self.outgoing_links(node):
                for in_port in linked_inputs:
                    yield in_port.node

    def input_neighbours(self, node: ToNode) -> Iterable[Node]:
        """Iterator over the input neighbours of a node.

        Args:
            node: Node to query.

        Returns:
            Iterator of nodes connected to `node` via incoming links.
            Nodes connected via multiple links will be returned multiple times.

        Examples:
            >>> df = dfg.Dfg()
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(0))
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(1))
            >>> list(df.hugr.input_neighbours(df.output_node))
            [Node(5), Node(5)]
        """
        return self.neighbours(node, Direction.INCOMING)

    def output_neighbours(self, node: ToNode) -> Iterable[Node]:
        """Iterator over the output neighbours of a node.

        Args:
            node: Node to query.

        Returns:
            Iterator of nodes connected to `node` via outgoing links.
            Nodes connected via multiple links will be returned multiple times.

        Examples:
            >>> df = dfg.Dfg()
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(0))
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(1))
            >>> list(df.hugr.output_neighbours(df.input_node))
            [Node(6), Node(6)]
        """
        return self.neighbours(node, Direction.OUTGOING)

    def num_incoming(self, node: Node) -> int:
        """The number of incoming links to a `node`.

        Examples:
            >>> df = dfg.Dfg()
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(0))
            >>> df.hugr.num_incoming(df.output_node)
            1
        """
        return sum(len(links) for (_, links) in self.incoming_links(node))

    def num_outgoing(self, node: ToNode) -> int:
        """The number of outgoing links from a `node`.

        Examples:
            >>> df = dfg.Dfg()
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(0))
            >>> df.hugr.num_outgoing(df.input_node)
            1
        """
        return sum(len(links) for (_, links) in self.outgoing_links(node))

    # TODO: num_links and _linked_ports

    def port_kind(self, port: InPort | OutPort) -> Kind:
        """The kind of a `port`.

        Examples:
            >>> df = dfg.Dfg(tys.Bool)
            >>> df.hugr.port_kind(df.input_node.out(0))
            ValueKind(Bool)
        """
        return self[port.node].op.port_kind(port)

    def port_type(self, port: InPort | OutPort) -> Type | None:
        """The type of a `port`, if the kind is
        :class:`ValueKind <hugr.tys.ValueKind>`, else None.

        Examples:
            >>> df = dfg.Dfg(tys.Bool)
            >>> df.hugr.port_type(df.input_node.out(0))
            Bool
        """
        op = self[port.node].op
        if isinstance(op, DataflowOp):
            return op.port_type(port)
        if isinstance(op, Call) and isinstance(port, OutPort):
            kind = self.port_kind(port)
            if isinstance(kind, ValueKind):
                return kind.ty
        return None

    def insert_hugr(self, hugr: Hugr, parent: ToNode | None = None) -> dict[Node, Node]:
        """Insert a HUGR entrypoint and all its descendants into this HUGR.

        If the inserted HUGR entrypoint was not its module root, some nodes will
        be ignored.

        Args:
            hugr: HUGR to insert.
            parent: Parent for root of inserted HUGR. Defaults to None.

        Returns:
            Mapping from node indices in inserted HUGR to their new indices
            in this HUGR.

        Examples:
            >>> d = dfg.Dfg()
            >>> h = Hugr()
            >>> h.insert_hugr(d.hugr)
            {Node(4): Node(1), Node(5): Node(2), Node(6): Node(3)}
        """
        mapping: dict[Node, Node] = {}

        for node in hugr.descendants():
            node_data = hugr[node]
            # relies on parents being inserted before any children
            if node == hugr.entrypoint:
                node_parent = parent
            else:
                if node_data.parent not in mapping:
                    raise ParentBeforeChild
                node_parent = mapping[node_data.parent]
            mapping[node] = self.add_node(
                node_data.op,
                node_parent,
                num_outs=node_data._num_outs,
                metadata=copy.copy(node_data.metadata),
            )

        for src, dst in hugr._links.items():
            # Neighbours that are not descendants of the entrypoint are ignored.
            if src.port.node not in mapping or dst.port.node not in mapping:
                continue
            self.add_link(
                mapping[src.port.node].out(src.port.offset),
                mapping[dst.port.node].inp(dst.port.offset),
            )
        return mapping

    def _overwrite_hugr(self, new_hugr: Hugr) -> None:
        """Modify a Hugr in place by replacing contents with those from a new Hugr."""
        self.module_root = new_hugr.module_root
        self.entrypoint = new_hugr.entrypoint
        self._nodes = new_hugr._nodes
        self._links = new_hugr._links
        self._free_nodes = new_hugr._free_nodes

    def _to_serial(self) -> SerialHugr:
        """Serialize the HUGR."""

        def _serialize_link(
            link: tuple[_SO, _SI],
        ) -> tuple[
            tuple[NodeIdx, PortOffset | None], tuple[NodeIdx, PortOffset | None]
        ]:
            src, dst = link
            s, d = self._constrain_offset(src.port), self._constrain_offset(dst.port)
            return (src.port.node.idx, s), (dst.port.node.idx, d)

        nodes: list[SerialOp] = []
        metadata = []
        entrypoint = 0
        for node_idx, data in enumerate(self._nodes):
            if data is None:
                continue
            node = Node(node_idx)
            serial_idx = len(nodes)

            # non contiguous indices will be erased
            nodes.append(data._to_serial(Node(serial_idx, NodeMetadata())))
            metadata.append(data.metadata.as_dict() if data.metadata else None)
            if self.entrypoint == node:
                entrypoint = serial_idx

        return SerialHugr(
            nodes=nodes,
            edges=[_serialize_link(link) for link in self._links.items()],
            metadata=metadata,
            entrypoint=entrypoint,
        )

    def _constrain_offset(self, p: P) -> PortOffset | None:
        """Constrain an offset to be a valid encoded port offset.

        Order edges should be encoded without an offset.
        """
        if p.offset < 0:
            assert p.offset == -1, "Only order edges are allowed with offset < 0"
            return None
        else:
            return p.offset

    def _connect_df_entrypoint_outputs(self) -> None:
        """If this Hugr was created by wrapping a dataflow operation entrypoint in a
        function, connect the entrypoint outputs to the function outputs.

        See `hugr.__init__` for more details.
        """
        from hugr.build import Function

        if not is_dataflow_op(self.entrypoint_op()):
            return

        func_node = self[self.entrypoint].parent
        assert func_node is not None, "Only module entrypoints may be HUGR roots"
        func_op = self[func_node].op
        if not isinstance(func_op, FuncDefn) or func_op._outputs is not None:
            return

        func = Function._new_existing(self, func_node)
        func.set_outputs(*self.entrypoint.outputs())

    @classmethod
    def _from_serial(cls, serial: SerialHugr) -> Hugr:
        """Load a HUGR from a serialized form."""
        assert serial.nodes, "The encoded Hugr is empty"

        def get_meta(idx: int) -> NodeMetadata:
            if serial.metadata and idx < len(serial.metadata):
                return NodeMetadata(serial.metadata[idx] or {})
            else:
                return NodeMetadata()

        # The first node is always the HUGR root.
        root_node = serial.nodes[0]
        assert (
            root_node.root.parent == 0
        ), "The encoded Hugr root must be the first node"
        root_op = root_node.root.deserialize()
        hugr = Hugr(root_op)
        hugr[hugr.entrypoint].metadata = get_meta(0)
        # In older versions, this may not be a module operation. If that's the
        # case, we rely on hugr's initializer to wrap the root in a module
        # and ignore the boilerplate nodes.
        boilerplate_nodes = hugr.num_nodes() - 1

        for idx, serial_node in enumerate(serial.nodes):
            if idx == 0:
                continue

            node_meta = get_meta(idx)
            parent: Node | None = Node(serial_node.root.parent)

            serial_node.root.parent = -1
            op = serial_node.root.deserialize()
            n = hugr._add_node(op, parent, metadata=node_meta, num_outs=op.num_out)
            assert (
                n.idx == idx + boilerplate_nodes
            ), "Nodes should be added contiguously"

            if idx == serial.entrypoint:
                hugr.entrypoint = n

        for (src_node, src_offset), (dst_node, dst_offset) in serial.edges:
            src = Node(src_node, _metadata=get_meta(src_node))
            dst = Node(dst_node, _metadata=get_meta(dst_node))
            if src_offset is None or dst_offset is None:
                src_op = hugr[src].op
                if isinstance(src_op, DataflowBlock | ExitBlock):
                    # Control flow edge
                    src_offset = 0
                    dst_offset = 0
                else:
                    # Order edge
                    hugr.add_order_link(src, dst)
                    continue
            hugr.add_link(
                src.out(src_offset),
                dst.inp(dst_offset),
            )

        return hugr

    @staticmethod
    def from_bytes(
        envelope: bytes, extensions: ExtensionRegistry | None = None
    ) -> Hugr:
        """Deserialize a byte string to a Hugr object.

        Some envelope formats can be read from a string. See :meth:`from_str`.

        Args:
            envelope: The byte string representing a Hugr envelope.
            extensions: If not None, an extension registry to resolve the custom
                operations and types.

        Returns:
            The deserialized Hugr object.

        Raises:
            ValueError: If the envelope does not contain exactly one module.
        """
        hugr = read_envelope_hugr(envelope)
        if extensions is not None:
            # TODO: This can be done during deserialization
            hugr.resolve_extensions(extensions)
        return hugr

    @staticmethod
    def from_str(envelope: str, extensions: ExtensionRegistry | None = None) -> Hugr:
        """Deserialize a string to a Hugr object.

        Not all envelope formats can be read from a string. See
        :meth:`from_bytes` for a more general method.

        Args:
            envelope: The string representing a Hugr envelope.
            extensions: If not None, an extension registry to resolve the custom
                operations and types.

        Returns:
            The deserialized Hugr object.

        Raises:
            ValueError: If the envelope does not contain exactly one module.
        """
        hugr = read_envelope_hugr_str(envelope)
        if extensions is not None:
            # TODO: This can be done during deserialization
            hugr.resolve_extensions(extensions)
        return hugr

    @staticmethod
    def from_model(module: model.Module) -> Hugr:
        """Import from the hugr model format."""
        from hugr.model.load import ModelImport

        loader = ModelImport(module=module)
        for i, node in enumerate(module.root.children):
            loader.import_node_in_module(node, i)
        loader.link_ports()
        loader.link_static_ports()
        loader.add_module_metadata()
        return loader.hugr

    def to_bytes(
        self,
        config: EnvelopeConfig | None = None,
        *,
        include_extensions: ExtensionRegistry | None = None,
    ) -> bytes:
        """Serialize the HUGR into an envelope byte string.

        Some envelope formats can be encoded into a string. See :meth:`to_str`.

        Args:
            config: The envelope configuration.
            include_extensions:
                If not None, the extensions to embed in the encoded envelope.
                If None, uses the registry returned by :meth:`used_extensions`.
                Standard extensions are ignored.
        """
        config = config or EnvelopeConfig.BINARY
        return _make_envelope(self.to_package(include_extensions), config)

    def to_str(
        self,
        config: EnvelopeConfig | None = None,
        *,
        include_extensions: ExtensionRegistry | None = None,
    ) -> str:
        """Serialize the package to a HUGR envelope string.

        Not all envelope formats can be encoded into a string.
        See :meth:`to_bytes` for a more general method.

        Args:
            config: The envelope configuration.
            include_extensions:
                If not None, the extensions to embed in the encoded envelope.
                If None, uses the registry returned by :meth:`used_extensions`.
                Standard extensions are ignored.
        """
        config = config or EnvelopeConfig.TEXT
        return _make_envelope_str(self.to_package(include_extensions), config)

    @deprecated("Use HUGR envelopes instead. See the `to_bytes` and `to_str` methods.")
    def to_json(self) -> str:
        """Serialize the HUGR to a JSON string.

        For most use cases, it is recommended to store a HUGR package instead.
        See :meth:`hugr.package.Package.to_bytes`.
        """
        return self._to_serial().to_json()

    def to_model(self) -> model.Module:
        """Export this module into the hugr model format."""
        from hugr.model.export import ModelExport

        export = ModelExport(self)
        region = export.export_region_module(self.module_root)
        return model.Module(region)

    def to_package(
        self, include_extensions: ExtensionRegistry | None = None
    ) -> Package:
        """Wrap this module into a Package.

        Args:
            include_extensions:
                If not None, the extensions to include in the package.
                If None, uses the registry returned by :meth:`used_extensions`.
                Standard extensions are ignored.
        """
        from hugr.package import Package
        from hugr.std import _std_extensions

        if include_extensions is None:
            include_extensions = self.used_extensions().used_extensions
        std = _std_extensions()
        extensions = [
            ext for ext in include_extensions.extensions.values() if ext.name not in std
        ]
        return Package(modules=[self], extensions=extensions)

    @classmethod
    @deprecated("Use HUGR envelopes instead. See the `to_bytes` and `to_str` methods.")
    def load_json(cls, json_str: str) -> Hugr:
        """Deserialize a JSON string into a HUGR.

        For most use cases, it is recommended to use package serialization instead.
        See :meth:`hugr.package.Package.from_bytes`.
        """
        json_dict = json.loads(json_str)
        serial = SerialHugr.load_json(json_dict)
        return cls._from_serial(serial)

    def render_dot(
        self, config: RenderConfig | None = None, root: Node | None = None
    ) -> gv.Digraph:
        """Render the HUGR to a graphviz Digraph.

        Args:
            config: Render configuration.
            root: Root node defining the set of nodes to render. By default this is the
                module root and all nodes are rendered. If this is a container node, all
                nodes under it are rendered. Every incoming edge to the rendered set and
                outgoing edge from it is also shown, with its other endpoint labelled
                with its node index.

        Returns:
            The graphviz Digraph.
        """
        from .render import DotRenderer

        return DotRenderer(config).render(self, root)

    def store_dot(
        self,
        filename: str,
        format: str = "svg",
        config: RenderConfig | None = None,
        root: Node | None = None,
    ) -> None:
        """Render the HUGR to a graphviz dot file.

        Args:
            filename: The file to render to.
            format: The format used for rendering ('pdf', 'png', etc.).
                Defaults to SVG.
            config: Render configuration.
            root: Root node defining the set of nodes to render. By default this is the
                module root and all nodes are rendered. If this is a container node, all
                nodes under it are rendered. Every incoming edge to the rendered set and
                outgoing edge from it is also shown, with its other endpoint labelled
                with its node index.
        """
        from .render import DotRenderer

        DotRenderer(config).store(self, filename=filename, format=format, root=root)

    def used_extensions(
        self, resolve_from: ext.ExtensionRegistry | None = None
    ) -> ext.ExtensionResolutionResult:
        """Get the extensions used by this HUGR, optionally resolving unresolved
        types and operations.

        This method modifies the HUGR in-place when resolve_from is provided,
        replacing Custom operations with ExtOp operations and opaque types with
        ExtType when their extensions are found in the registry.

        Args:
            resolve_from: Optional extension registry to resolve against.
                If None, opaque types and Custom ops will not be resolved.

        Returns:
            The result of resolving the extensions, containing the used
            extensions and a list of referenced but unresolved extensions.

        Example:
            >>> from hugr.build import Dfg
            >>> Dfg(tys.Qubit).hugr.used_extensions().ids()
            {'prelude'}
        """
        from hugr.ext import ExtensionResolutionResult

        result = ExtensionResolutionResult()

        for node in self:
            op = self[node].op
            # _resolve_used_extensions returns the resolved op and the extensions
            resolved_op, op_result = op._resolve_used_extensions(resolve_from)
            self[node].op = resolved_op
            result.extend(op_result)

        result._extend_with_transitive_ops(resolve_from)

        return result

    def resolve_extensions(self, registry: ext.ExtensionRegistry) -> Hugr:
        """Resolve extension references in the types and operations of the HUGR.

        This is an alias for :meth:`used_extensions` that discards the computed
        extensions.

        Args:
            registry: The extension registry to resolve against.
        """
        self.used_extensions(resolve_from=registry)
        return self
